import ast
import os

from py10x_kernel import OsUser

from core_10x.global_cache import cache
from core_10x.rc import RC
from core_10x.xdate_time import XDateTime, date, datetime
from core_10x.exec_control import GRAPH_ON, GRAPH_OFF
#from core_10x.resource import Resource

#===================================================================================================================================
#   date_format: str    = default_value
#
#   def date_format_apply(cls, fmt: str):
#     XDateTime.set_default_format(fmt)
#
#   datetime_format: str
#
#   def datetime_format_get(cls):
#       return f'{cls.date_format} %H:%M%S'
#
#   def datetime_format_apply(cls, value):
#       ...
#===================================================================================================================================


class classproperty(property):
    def __get__(self, obj, objtype: type = None):
        return self.fget(objtype)


class _EnvVars:
    class Var:
        def __init__(self, env_var_class, attr_name: str, value = None):
            self.env_var_class = env_var_class
            self.attr_name = attr_name
            if value is None:
                value = getattr(env_var_class, attr_name)
            self.value = value

        def __bool__(self):
            return bool(self.value)

        def check(self, f = None, err = 'is not defined'):
            value = self.value
            rc = bool(value) if not f else f(value)
            if not rc:
                raise ValueError(f'{self.env_var_class.var_name(self.attr_name)} {err}')

            return value

    s_converters = {
        bool:       lambda s: bool(ast.literal_eval(s.lower().capitalize())),
        int:        lambda s: int(ast.literal_eval(s)),
        float:      lambda s: float(ast.literal_eval(s)),
        str:        lambda s: s,
        date:       lambda s: XDateTime.str_to_date(s),
        datetime:   lambda s: XDateTime.str_to_datetime(s),
        #Resource:   lambda s: Resource
    }

    @classmethod
    def _getter(cls, data_type, var_name: str, f_get, f_apply):
        str_value = os.getenv(var_name)
        if str_value is not None:
            f_convert = cls.s_converters.get(data_type)
            assert f_convert, f'Unknown data type {data_type}'
            try:
                value = f_convert(str_value)
            except Exception as e:
                raise TypeError(f'Variable {var_name} - could not convert {str_value} to {data_type}') from e
        else:
            try:
                value = f_get(cls)
            except Exception as e:
                rc = RC(False)  # -- capture the exc
                raise RuntimeError(f'{cls}.{var_name} - failed while getting a value\n{rc.error()}') from e

        if f_apply:
            try:
                f_apply.__get__(cls)(value)
            except Exception as e:
                rc = RC(False)  #-- capture the exc
                raise ValueError(f'{cls}.{var_name} - failed while applying value: {value}\n{rc.error()}') from e

        return value

    @classmethod
    def full_getter(cls, data_type, var_name: str, f_get, f_apply):
        f = lambda cls: cls._getter(data_type, var_name, f_get, f_apply)
        # return classmethod(cache(f))
        return cache(f)

    @classmethod
    def create_var_name(cls, env_name: str, attr_name: str) -> str:
        return f'{env_name}_{attr_name.upper()}'

    @classmethod
    def var_name(cls, var_or_name) -> str:
        if isinstance(var_or_name, cls.Var):
            var_or_name = var_or_name.attr_name
        return cls.create_var_name(cls.s_env_name, var_or_name)

    s_env_name: str = None
    var = None
    def __init_subclass__(cls, env_name: str = None, **kwargs):
        assert env_name, 'env_name is required'

        cls.s_env_name = env_name
        annotations = cls.__annotations__
        assert annotations, 'No variables are defined'

        cls_dict = cls.__dict__
        for name, data_type in annotations.items():
            default_value = cls_dict.get(name)
            if default_value is None:  # -- no default value, let's locate the getter
                f_get_name = f'{name}_get'
                f_get = cls_dict.get(f_get_name)
                assert f_get, f'Variable {name} must define either a default value or a getter {f_get_name}(cls)'
                # -- TODO: check signature: f_get(cls)
            else:
                f_get = lambda cls, def_value = default_value: def_value

            f_apply_name = f'{name}_apply'
            f_apply = cls_dict.get(f_apply_name)  # -- f(cls, value)

            var_name = cls.create_var_name(env_name, name)
            full_getter = cls.full_getter(data_type, var_name, f_get, f_apply)
            setattr(cls, name, classproperty(full_getter))

            cls.var = cls.AccessVar(cls)

    class AccessVar:
        def __init__(self, env_vars_class):
            self.env_vars_class = env_vars_class

        def __getattr__(self, item):
            cls = self.env_vars_class
            value = getattr(cls, item)
            if value is None:
                raise AssertionError(f'Unknown var name {item}')

            return _EnvVars.Var(cls, item, value)

class EnvVars(_EnvVars, env_name='XX'):
    build_area: str
    parent_build_area: str          = 'dev'

    master_password_key: str        = 'XX_MASTER_PASSWORD'

    # examples_ts_store_uri: str    = ''            #-- e.g., mongodb://localhost/examples
    vault_ts_store_uri: str         = ''            #-- to store/auto retrieve security credentials for each user/resource; default is main_ts_store_uri
    main_ts_store_uri: str          = ''            #-- e.g., 'mongodb://localhost:27018/main'
    log_ts_store_uri: str           = ''            #-- to keep all app logs in a specific store, e.g., 'mongodb://localhost/app_logs'
    use_ts_store_per_class: bool    = True          #-- use TsStore per Traitable class associations
    functional_account_prefix: str  = 'xx'          #-- used in usernames to distinguish a regular username from a functional account

    graph_on: bool                  = False         #-- whether GRAPH is ON by default
    use_ts_store_transactions: bool = False         #-- TsStore transactions, e.g. when saving history

    date_format: str = XDateTime.FORMAT_ISO

    sdlc_area: str

    def vault_ts_store_uri_get(self) -> str:
        return self.main_ts_store_uri

    def build_area_get(self) -> str:
        return OsUser.me.name()

    @classmethod
    def date_format_apply(cls, value):
        XDateTime.set_default_format(value)

    def sdlc_area_get(self) -> str:
        ba = self.build_area
        pba = self.parent_build_area
        return f'{ba}/{pba}' if pba else ba



from core_10x.exec_control import CACHE_ONLY
from core_10x.traitable import T, Traitable, Ui

COLORS = ('black', 'white', 'green', 'lightgreen', 'red', 'blue', 'grey')

class StyleSheet(Traitable):
    foreground: str     = T('black',        ui_hint = Ui(widget_type = Ui.WIDGET_TYPE.CHOICE))
    background: str     = T('white',        ui_hint = Ui(widget_type = Ui.WIDGET_TYPE.CHOICE, flags = Ui.SEPARATOR))

    font: str           = T('Helvetica',    ui_hint = Ui(widget_type = Ui.WIDGET_TYPE.CHOICE))
    font_style: bool    = T(False,          ui_hint = Ui('italic', right_label = True))
    font_weight: bool   = T(False,          ui_hint = Ui('bold', flags = Ui.SEPARATOR, right_label = True))

    border_style: bool  = T(False)
    border_color: str   = T('blue',         ui_hint = Ui(widget_type = Ui.WIDGET_TYPE.CHOICE))
    border_width: int   = T(2,              ui_hint = Ui(flags = Ui.SEPARATOR))

    show_me: str        = T('This is how it will look...',  ui_hint = Ui('WYSIWYG', min_width = 50))

    def foreground_choices(self, t):    return COLORS
    def background_choices(self, t):    return COLORS
    def border_color_choices(self, t):  return COLORS
    def font_choices(self, t):          return ('Times New Roman', 'Helvetica', 'Courier New')

    def show_me_style_sheet(self) -> dict:
        ss = {
            Ui.FG_COLOR:        self.foreground,
            Ui.BG_COLOR:        self.background,
            Ui.FONT:            self.font,
            Ui.FONT_STYLE:      'italic'   if self.font_style    else 'normal',
            Ui.FONT_WEIGHT:     'bold'     if self.font_weight   else 'normal',
            Ui.BORDER_WIDTH:    f'{self.border_width}px',
            Ui.BORDER_STYLE:    'solid'    if self.border_style  else '',
            Ui.BORDER_COLOR:    self.border_color,
        }
        print(ss)
        return ss

if __name__ == '__main__':
    from ui_10x.traitable_editor import TraitableEditor

    with CACHE_ONLY():
        sheet = StyleSheet()
        e = TraitableEditor(sheet, _confirm = True)
        e.dialog().exec()





from __future__ import annotations

from functools import partial
from unittest.mock import MagicMock

import pytest
from core_10x.exec_control import BTP, INTERACTIVE
from core_10x.rc import RC
from core_10x.trait_method_error import TraitMethodError
from core_10x.traitable import Traitable


class X(Traitable):
    x: int
    y: int
    z: int

    def x_set(self, trait, value) -> RC:
        self.raw_set_value(trait, value)
        return self.set_values(y=value)

    def z_get(self):
        return self.x


def callback(btp, x, t, v):
    assert btp is BTP.current()
    # assert x.get_value(t) == v
    print(btp, BTP.current())
    x.bui_class().update_ui_node(x, t)


def test_ui_nodes():
    x = X(x=1)

    def t(ov, v):
        mx = MagicMock(side_effect=partial(callback, BTP.current(), x, x.T.x, v))
        my = MagicMock(side_effect=partial(callback, BTP.current(), x, x.T.y, v))
        mz = MagicMock(side_effect=partial(callback, BTP.current(), x, x.T.z, v))
        x.bui_class().create_ui_node(x, x.T.x, mx)
        x.bui_class().create_ui_node(x, x.T.y, my)
        x.bui_class().create_ui_node(x, x.T.z, mz)

        assert x.x == x.y == x.z == ov

        x.x = v

        assert mx.call_count == my.call_count == mz.call_count == 1
        assert x.x == x.y == x.z == v

        return mx, my, mz

    with INTERACTIVE() as i0:
        print(i0)
        mx, my, mz = t(1, 2)
        with INTERACTIVE() as i:
            print(i)
            t(2, 3)
        i.export_nodes()
        assert x.x == x.y == x.z == 3
        assert mx.call_count == my.call_count == mz.call_count == 2


def test_exception():
    def callback():
        x.bui_class().update_ui_node(x, x.T.x)  # TODO: should not be required when throwing
        raise RuntimeError('test')

    x = X(x=1)
    with INTERACTIVE():
        m = MagicMock(side_effect=callback)
        x.bui_class().create_ui_node(x, x.T.x, m)
        with pytest.raises(TraitMethodError):  # TODO: should be throwing RuntimeError?
            x.x = 2
        assert x.x == 2  # TODO: should not set value when callback failed?
        with INTERACTIVE() as i:
            x.x = 3
        with pytest.raises(RuntimeError):  # TODO: RuntimeError or TraitMethodError?
            i.export_nodes()

        assert m.call_count == 2



#TODO: delete this file, it is duplicated in ui_10x.utils
assert False, "use ui_10x.utils"
# from datetime import date
# import inspect
# import platform
#
# from PyQt6.QtWidgets import QLabel, QWidget, QCalendarWidget, QMessageBox, QGroupBox, QButtonGroup, QRadioButton, QApplication
# from PyQt6.QtWidgets import QListWidget, QLineEdit, QTreeWidgetItem, QTreeWidget, QSizePolicy
# from PyQt6.QtWidgets import QFrame, QVBoxLayout, QHBoxLayout, QSplitter, QDialog, QScrollArea, QPushButton, QStyle
# from PyQt6.QtCore import Qt, QObject, QThread, pyqtSignal, QBuffer, QIODevice    #, QByteArray
# from PyQt6.QtGui import QColor, QGuiApplication, QPixmap
#
# # from asu.nx.directory import Directory
#
# from core_10x.named_constant import NamedConstant
# from core_10x.rc import RC
# from core_10x.global_cache import cache, singleton
#
# class UxAsync(QObject):
#     SIGNAL          = pyqtSignal()
#     s_instances     = {}
#
#     @classmethod
#     def _check_thread(cls, bound_method):
#         if QThread.currentThread() != QGuiApplication.instance().thread():
#             raise RuntimeError(f'You must have called UxAsync.init({bound_method.__name__}) in the UI thread')
#
#     @classmethod
#     def init(cls, bound_method):
#         instance = cls.s_instances.get(bound_method)
#         if not instance:
#             cls._check_thread(bound_method)
#             instance = cls()
#             cls.s_instances[bound_method] = instance
#
#             instance.SIGNAL.connect(bound_method, type = Qt.ConnectionType.QueuedConnection)
#
#         return instance
#
#     @classmethod
#     def call(cls, bound_method):
#         instance = cls.init(bound_method)
#         instance.SIGNAL.emit()
#
# class UxRadioBox(QGroupBox):
#     def __init__(self, named_constant_class, title = '', horizontal = False, default_value: NamedConstant = None):
#         assert inspect.isclass(named_constant_class) and issubclass(named_constant_class, NamedConstant), 'subclass of NamedConstant is expected'
#
#         self.nm_class = named_constant_class
#         self.values = values = list(named_constant_class.s_dir.values())
#
#         if default_value is None:
#             default = 0
#         else:
#             assert isinstance(default_value, named_constant_class), f'{named_constant_class} - unknown default_values = {default_value}'
#             for i, c_def in enumerate(values):
#                 if c_def is default_value:
#                     default = i
#                     break
#             else:
#                 assert False, f'{default_value} is not found'
#
#         if title:
#             super().__init__(title)
#         else:
#             super().__init__()
#
#
#         lay = QHBoxLayout() if horizontal else QVBoxLayout()
#         self.setLayout(lay)
#
#         self.group = group = QButtonGroup()
#         for i, c_def in enumerate(named_constant_class.s_dir.values()):
#             rb = QRadioButton(c_def.label)
#             lay.addWidget(rb)
#             group.addButton(rb, id = i)
#         group.button(default).setChecked(True)
#
#     def choice(self):
#         i = self.group.checkedId()
#         return self.values[i]
#
# def uxSeparator(horizontal = True) -> QLabel:
#     direction = QFrame.Shape.HLine if horizontal else QFrame.Shape.VLine
#     sep = QLabel()
#     sep.setFrameStyle(direction | QFrame.Shadow.Sunken)
#     return sep
#
# def uxToClipboard(text: str, **kwargs):
#     cp = QGuiApplication.clipboard()
#     cp.setText(text)
#
# def uxFromClipboard(**kwargs) -> str:
#     cp = QGuiApplication.clipboard()
#     return cp.text()
#
# def uxSuccess(text: str, parent = None, title = ''):
#     if not title:
#         title = 'Success'
#     if not parent:
#         parent = None
#     QMessageBox.information(parent, title, text)
#
# def uxWarning(text: str, parent = None, title = ''):
#     if not title:
#         title = 'Warning'
#     if not parent:
#         parent = None
#     QMessageBox.warning(parent, title, text)
#
# def uxAnswer(question: str, parent = None, title = '') -> bool:
#     if not title:
#         title = 'Waiting for your answer...'
#     if not parent:
#         parent = None
#     sb = QMessageBox.question(parent, title, question)
#     return True if sb == QMessageBox.StandardButton.Yes else False
#
# def uxMultiChoiceAnswer(named_constant_class, parent = None, title = '', default_value: NamedConstant = None):
#     if not title:
#         title = 'Pick one of the choices below:'
#
#     if not parent:
#         parent = None
#
#     box = UxRadioBox(named_constant_class, default_value = default_value)
#     d = UxDialog(box, parent = parent, title = title, cancel = '')
#     d.exec()
#     return box.choice()
#
# def uxPushButton(label: str, callback = None, style_icon = None, flat = False):
#     if style_icon:
#         if isinstance(style_icon, str):
#             style_icon = getattr(QStyle, 'SP_' + style_icon, None)
#             assert style_icon, f"Unknown style_icon = '{style_icon}'"
#
#         assert isinstance(style_icon, int), 'Currently only str or int are supported for style_icon'
#
#         style = QApplication.style()
#         icon = style.standardIcon(style_icon)
#         button = QPushButton(icon, label)
#     else:
#         button = QPushButton(label)
#
#     if callback:
#         button.clicked.connect(callback)
#
#     if flat:
#         button.setFlat(True)
#
#     return button
#
# class UxDialog(QDialog):
#     def _createButton(self, ok: bool, button_spec) -> QPushButton:
#         if not button_spec:
#             return None
#
#         if isinstance(button_spec, str):
#             label = button_spec
#             icon = ''
#
#         elif isinstance(button_spec, tuple):
#             try:
#                 label, icon = button_spec
#             except Exception:
#                 return None
#
#         else:
#             return None
#
#         cb = self.onOk if ok else self.onCancel
#         return uxPushButton(label, callback = cb, style_icon = icon)
#
#     def __init__(
#             self,
#             w: QWidget,
#             parent: QWidget     = None,
#             title: str          = None,
#             ok                  = 'Ok',
#             cancel              = 'Cancel',
#             accept_callback     = None,
#             cancel_callback     = None,
#             min_width           = 0,
#             min_height          = 0,
#             window_flags        = None
#     ):
#         """
#         :param w:               a QWidget to show
#         :param parent:          parent widget, if any
#         :param title:           dialog title
#         :param accept_callback: context.method, where: method( context ) -> RC
#         :param ok, cancel:      labels and corresponding roles for Ok and Cancel buttons. No button if empty
#         """
#         super().__init__( parent )
#         if title:
#             self.setWindowTitle( title )
#
#         if window_flags is None:
#             window_flags = Qt.WindowType.CustomizeWindowHint | Qt.WindowType.Window | Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint
#
#         self.setWindowFlags(window_flags)
#         self.accept_callback = accept_callback
#         self.cancel_callback = cancel_callback if cancel_callback else self.reject
#
#         if ok:
#             ok = self._createButton(True, ok)
#         if cancel:
#             cancel = self._createButton(False, cancel)
#
#         lay = QVBoxLayout()
#         self.setLayout(lay)
#
#         lay.addWidget(w)
#
#         self.w_message = QLabel()
#         self.w_message.setStyleSheet('color: red;')
#         lay.addWidget(self.w_message)
#
#         if ok or cancel:
#             lay.addWidget(uxSeparator())
#
#             bar = QHBoxLayout()
#             lay.addLayout(bar)
#
#             bar.addWidget(QLabel(), stretch = 1)
#             if ok:
#                 bar.addWidget(ok)
#             if cancel:
#                 bar.addWidget(cancel)
#
#         if min_width > 0:
#             self.setMinimumWidth(min_width)
#         if min_height > 0:
#             self.setMinimumHeight(min_height)
#
#     def onOk(self):
#         if self.accept_callback:
#             rc = self.accept_callback()
#             if not rc:
#                 self.message(rc.err())
#                 return
#
#         self.done(1)
#
#     def onCancel(self):
#         self.reject()
#         self.done(0)
#
#     def message(self, text: str):
#         self.w_message.setText(text)
#
# def uxPickDate(title = 'Pick a Date', show_date: date = None, grid = True, default = None):
#     cal = QCalendarWidget()
#     cal.setGridVisible(bool(grid))
#     if show_date:
#         cal.setSelectedDate(show_date)
#     dlg = UxDialog(cal, title = title)
#     rc = dlg.exec()
#
#     return cal.selectedDate().toPyDate() if rc else default
#
# class UxStyleSheet:
#     @classmethod
#     def dumps(cls, data: dict) -> str:
#         return '\n'.join(f'{name}: {value};' for name, value in data.items())
#
#     @classmethod
#     def loads(cls, sheet: str) -> dict:
#         res = {}
#         pairs = sheet.split(';')
#         for pair in pairs:
#             pair = pair.strip()
#             if pair:
#                 name_value = pair.split(':')
#                 if len(name_value) == 2:
#                     name = name_value[0].strip()
#                     value = name_value[1].strip()
#                     res[name] = value
#
#         return res
#
#     @classmethod
#     def slice(cls, w: QWidget, *attr_names) -> list:
#         data = cls.loads(w.styleSheet())
#         return [ data.get(name) for name in attr_names ]
#
#     @classmethod
#     def update(cls, w: QWidget, sheet_update: str, replace = False):
#         if not replace:
#             sh = w.styleSheet()
#             data = cls.loads(sh)
#             update_data = cls.loads(sheet_update)
#             data.update(update_data)
#             new_sh = cls.dumps(data)
#         else:
#             new_sh = sheet_update
#
#         w.setStyleSheet(new_sh)
#
#     @classmethod
#     def _color(cls, style_sheet: str, attr_name: str):
#         data = cls.loads(style_sheet)
#         str_color = data.get(attr_name)
#         return QColor(str_color) if str_color else None
#
#     @classmethod
#     def fg_color(cls, style_sheet: str) -> QColor:
#         return cls._color(style_sheet, 'color')
#
#     @classmethod
#     def bg_color(cls, style_sheet: str) -> QColor:
#         return cls._color(style_sheet, 'background-color')
#
#     @classmethod
#     def color(cls, style_sheet: str, background = False) -> QColor:
#         return cls.bg_color(style_sheet) if background else cls.fg_color(style_sheet)
#
# s_verticalAlignmentMap = {
#     -1:     Qt.AlignmentFlag.AlignTop,
#     0:      Qt.AlignmentFlag.AlignVCenter,
#     1:      Qt.AlignmentFlag.AlignBottom,
# }
# s_horizontalAlignmentMap = {
#     -1:     Qt.AlignmentFlag.AlignLeft,
#     0:      Qt.AlignmentFlag.AlignCenter,
#     1:      Qt.AlignmentFlag.AlignRight,
# }
# def uxTextAlignment(align_value: int, horizontal = True) -> int:
#     if horizontal:
#         return s_horizontalAlignmentMap.get(align_value, Qt.AlignmentFlag.AlignRight) | Qt.AlignmentFlag.AlignVCenter
#
#     raise NotImplementedError
#
# class UX_SCROLL:
#     OFF         = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
#     ON          = Qt.ScrollBarPolicy.ScrollBarAlwaysOn
#     AS_NEEDED   = Qt.ScrollBarPolicy.ScrollBarAsNeeded
#
# def uxMakeScrollable(w: QWidget, h = UX_SCROLL.AS_NEEDED, v = UX_SCROLL.AS_NEEDED) -> QWidget:
#     if h is UX_SCROLL.OFF and v is UX_SCROLL.OFF:
#         return w
#
#     sa = QScrollArea()
#     sa.setWidget(w)
#     sa.setHorizontalScrollBarPolicy(h)
#     sa.setVerticalScrollBarPolicy(v)
#     return sa
#
# class UxSearchableList(QGroupBox):
#     def __init__(
#         self,
#         text_widget: QLineEdit = None,
#         reset_selection = True,
#         choices: list   = None,
#         select_hook     = None,
#         sort            = False,
#         case_sensitive  = False,
#         title           = ''
#     ):
#         """
#         :param text_widget: if an external text_widget given, it is used, otherwise its own is created on top of the list
#         :param reset_selection: if True, text_widget will be reset on new selection
#         :param choices: all choices
#         :param select_hook: if given, will be called after a new selection is made: select_hook( selected_choice: str )
#         :param sort: sort choices if True
#         :param case_sensitive: whether to search with case sensitivity
#         :param title: an optional title
#         """
#         super().__init__()
#         self.m_sort = sort
#         self.m_initialChoices = choices if not sort else sorted( choices )
#         self.m_currentChoices = self.m_initialChoices
#         self.m_hook = select_hook
#         self.m_resetSelection = reset_selection
#         self.m_caseSensitive = case_sensitive
#         self.m_choice = None
#
#         if title:
#             self.setTitle( title)
#         lay = QVBoxLayout()
#
#         if text_widget:
#             self.m_field = text_widget
#         else:
#             self.m_field = QLineEdit()
#             lay.addWidget( self.m_field )
#         self.m_field.textEdited.connect( self.processInput )
#
#         self.m_list = QListWidget()
#
#         self.m_list.addItems( self.m_initialChoices )
#         self.m_list.clicked.connect( self.itemSelected )
#         lay.addWidget( self.m_list )
#
#         self.setLayout( lay )
#
#     def addChoice( self, choice: str ):
#         if not choice in self.m_initialChoices:
#             self.m_initialChoices.append( choice )
#             if self.m_sort:
#                 self.m_initialChoices = sorted( self.m_initialChoices )
#
#             self.m_list.clear()
#             self.m_list.addItems( self.m_initialChoices )
#
#     def removeChoice( self, choice: str ):
#         if choice in self.m_initialChoices:
#             self.m_initialChoices.remove( choice )
#             self.m_list.clear()
#             self.m_list.addItems( self.m_initialChoices )
#
#     def processInput( self, input ):
#         self.m_list.clear()
#         if not input:
#             self.m_currentChoices = self.m_initialChoices
#         else:
#             if not self.m_currentChoices:
#                 self.m_currentChoices = self.m_initialChoices
#
#             if not self.m_caseSensitive:
#                 input = input.lower()
#                 self.m_currentChoices = [ s for s in self.m_currentChoices if not s.lower().find( input ) == -1 ]
#             else:
#                 self.m_currentChoices = [ s for s in self.m_currentChoices if not s.find( input ) == -1 ]
#
#         self.m_list.addItems( self.m_currentChoices )
#
#     def itemSelected( self, item ):
#         i = item.row()
#         text = self.m_currentChoices[ i ]
#         self.reset()
#
#         items = self.m_list.findItems( text, Qt.MatchExactly )
#         if len( items ):
#             items[ 0 ].setSelected( True )
#             text_s = '' if self.m_resetSelection else text
#             self.m_field.setText( text_s )
#
#         self.m_choice = text
#         if self.m_hook:
#             self.m_hook( text )
#
#     def choice( self ) -> str:
#         return self.m_choice
#
#     def reset( self ):
#         if not self.m_currentChoices == self.m_initialChoices:
#             self.m_currentChoices = self.m_initialChoices
#             self.m_list.clear()
#             self.m_list.addItems( self.m_initialChoices )
#
# # class UxTreeViewer( QTreeWidget ):
# #     s_labelMaxLength    = 40
# #
# #     def __init__( self, dir: Directory, select_hook = None, label_max_length = -1, expand = False, **kwargs ):
# #         super().__init__()
# #         if label_max_length < 0:
# #             label_max_length = self.s_labelMaxLength
# #
# #         self.m_dir = dir
# #         self.m_selectHook = select_hook
# #
# #         dir_value = dir.dir_value()
# #         dir_name = dir.dir_name()
# #         if dir_value and dir_name and dir_name != dir_value:
# #             num_cols = 2
# #             header_labels = [ dir.dir_name(), 'Description' ]
# #         else:
# #             num_cols = 1
# #             header_labels = [ dir.show_value() ]
# #
# #         self.m_numCols = num_cols
# #         self.setColumnCount( num_cols )
# #         self.setHeaderLabels( header_labels )
# #         self.itemClicked.connect( self.typeTreeItemClicked )
# #
# #         for subdir in dir.dir_members().values():  # -- without the root!
# #             self.createTree( subdir, self, label_max_length, num_cols )
# #
# #         if expand:
# #             for i in range( self.topLevelItemCount() ):
# #                 top = self.topLevelItem( i )
# #                 top.setExpanded( True )
# #
# #         self.resizeColumnToContents( 0 )
# #         if num_cols == 2:
# #             self.resizeColumnToContents( 1 )
# #         self.setSizePolicy( QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding )
# #
# #     class TreeItem( QTreeWidgetItem ):
# #         def __init__( self, parent_node, dir: Directory ):
# #             super().__init__( parent_node )
# #             self.m_dir = dir
# #
# #         def dir( self ):    return self.m_dir
# #
# #     @classmethod
# #     def createTree( cls, dir: Directory, parent_node, label_max_length: int, num_cols: int ):
# #         node = cls.TreeItem( parent_node, dir )
# #         show_value = dir.show_value()
# #         node.setText( 0, show_value )
# #         if num_cols == 2:
# #             label = dir.dir_name()
# #             if label and label != show_value:
# #                 if len( label ) > label_max_length:
# #                     label = label[ :label_max_length ] + '...'
# #
# #                 node.setText( 1, label )
# #
# #         node.setToolTip( num_cols - 1, dir.dir_name() )
# #
# #         for subdir in dir.dir_members().values():
# #             cls.createTree( subdir, node, label_max_length, num_cols )
# #
# #     def typeTreeItemClicked( self, item: QTreeWidgetItem ):
# #         self.onTagSelected( item.dir() )
# #
# #     def onTagSelected( self, dir: Directory ):
# #         if self.m_selectHook:
# #             self.m_selectHook( dir )
#
# # class UxPixmap:
# #     COLOR_AUTO      = Qt.ColorScheme.AutoColor
# #     COLOR_DITHER    = Qt.ColorScheme.ColorOnly
# #     COLOR_MONO      = Qt.ColorScheme.MonoOnly
# #     RATIO_IGNORE    = Qt.ColorScheme.IgnoreAspectRatio
# #     RATIO_KEEP      = Qt.ColorScheme.KeepAspectRatio
# #     RATIO_KEEP_EXP  = Qt.ColorScheme.KeepAspectRatioByExpanding
# #
# #     s_sourceTypeMap = {
# #         str:        QPixmap.load,
# #         bytes:      QPixmap.loadFromData,
# #         bytearray:  QPixmap.loadFromData,
# #     }
# #     def __init__( self, source = None, image_format = '', color_flags = COLOR_AUTO ):
# #         self.m_pixmap = QPixmap()
# #         rc = self.load( source, image_format = image_format, color_flags = color_flags )
# #         if not rc:
# #             raise RuntimeError( rc.err() )
# #
# #     def load( self, source, image_format = '', color_flags = COLOR_AUTO ) -> RC:
# #         if not source:
# #             return RC( True )
# #
# #         method = self.s_sourceTypeMap.get( type( source ) )
# #         if not method:
# #             return RC( False, f'UxPixmap - unknown type of source = {type( source )}' )
# #
# #         rc = method( self.m_pixmap, source, format = image_format, flags = color_flags )
# #         if not rc:
# #             return RC( False, f'UxPixmap - failed to initialize from source = {source}' )
# #
# #         return RC( True )
# #
# #     def _toData( self, array: bytearray, format = '', quality = -1 ) -> bool:
# #         buffer = QBuffer( array )
# #         buffer.open( QIODevice.WriteOnly )
# #         return self.m_pixmap.save( buffer, format = format, quality = quality )
# #
# #     s_destinationTypeMap = {
# #         str:        QPixmap.save,
# #         bytearray:  _toData,
# #     }
# #     def save( self, destination, color_format = '', quality = -1 ) -> RC:
# #         method = self.s_destinationTypeMap.get( type( destination ) )
# #         if not method:
# #             return RC( False, f'UxPixmap - unknown destination type = {type( destination )}' )
# #
# #         rc = method( self, destination, format = color_format, quality = quality )
# #         if not rc:
# #             return RC( False, f'UxPixmap - failed to save to destination = {destination}' )
# #
# #         return RC( True )
# #
# #     def render( self, label: QLabel, w = 0, h = 0, scaled = RATIO_IGNORE, smooth = True ):
# #         transform_mode = Qt.SmoothTransformation if smooth else Qt.FastTransformation
# #         pixmap = self.m_pixmap
# #         if w > 0 and h > 0:
# #             pixmap = pixmap.scaled( w, h, aspectRatioMode = scaled, transformMode = transform_mode )
# #         label.setPixmap( pixmap )
#
# @singleton
# class UxClipBoard:
#     def __init__(self):
#         self.dir = {}
#         self.entity = None
#         self.trait_name = ''
#
#     # def copy(self, tag: str, value):
#     #     self.dir[tag] = value
#     #
#     # def paste(self, tag: str):
#     #     return self.dir.get(tag)
#     #
#     # def clear(self):
#     #     self.dir = {}
#     #
#     # def default_tag( self ) -> str:
#     #     assert self.entity and self.trait_name, 'entity and trait name are not defined'
#     #
#     #     cls = self.entity.__class__
#     #     trait = cls.trait(self.trait_name)
#     #     assert trait, f"{cls}: unknown trait '{self.trait_name}'"
#     #     #tag_label = trait.label if trait.label else self.trait_name
#     #     tag_label = self.trait_name
#     #     return f"{self.entity}: '{tag_label}'"
#     #
#     # def layout(self) -> QWidget:
#     #     w = QWidget()
#     #     lay = QVBoxLayout()
#     #     w.setLayout(lay)
#     #
#     #     add_row = QHBoxLayout()
#     #     add_b = uxPushButton('Add', callback = self.onCopy)
#     #     self.m_addNameWidget = add_name = QLineEdit()
#     #     add_name.setText(self.default_tag())
#     #     add_row.addWidget(add_b)
#     #     add_row.addWidget(add_name)
#     #     lay.addLayout(add_row)
#     #
#     #     lay.addWidget(uxSeparator())
#     #
#     #     self.m_sl = sl = UxSearchableList(choices = list( self.dir.keys() ), select_hook = self.onPaste, title = 'Use Data')
#     #     lay.addWidget(sl)
#     #
#     #     return w
#     #
#     # def popup(self, parent_widget: QWidget, entity, trait_name: str):
#     #     self.m_parent = parent_widget
#     #     self.entity = entity
#     #     self.trait_name = trait_name
#     #     self.m_d = d = UxDialog(self.layout(), parent = parent_widget, cancel = '', title = 'Copy or Paste Data')
#     #     d.exec()
#     #
#     # def onCopy(self):
#     #     name = self.m_addNameWidget.text()
#     #     if name:
#     #         value = self.entity.get_value( self.trait_name )
#     #         if value is not None:
#     #             self.copy( name, value )
#     #
#     #     self.m_d.close()
#     #
#     # def onPaste( self, name: str ):
#     #     value = self.paste( name )
#     #     if value is not None:
#     #         rc = self.m_entity.setValue( self.m_traitName, value )
#     #         if not rc:
#     #             uxWarning(
#     #                 f"Data '{name}' is not suitable for '{self.m_entity.trait( self.m_traitName ).m_label}'",
#     #                 parent = self.m_parent,
#     #                 title = 'Invalid Data'
#     #             )
#     #
#     #     self.m_d.close()
#
# @cache
# def uxInit(style = '') -> QApplication:
#     if not style:
#        style = platform.system()
#
#     app = QApplication([])
#     app.setStyle(style)
#     return app
#



from __future__ import annotations

import ast
import copy
import functools
import inspect
import locale
import platform
import sys
from inspect import Parameter
from types import GenericAlias
from typing import get_origin, get_type_hints

from py10x_kernel import BTrait

from core_10x.named_constant import NamedConstant, NamedCallable
from core_10x.package_refactoring import PackageRefactoring
from core_10x.rc import RC
from core_10x.trait_definition import T, TraitDefinition, Ui
from core_10x.xnone import XNone


class TraitMetaclass(type(BTrait)):
    def __new__(cls, name, bases, class_dict, **kwargs):
        if '__slots__' not in class_dict:
            class_dict['__slots__'] = ()
        return super().__new__(cls, name, bases, class_dict, **kwargs)


class Trait(BTrait, metaclass=TraitMetaclass):
    __slots__ = ('fmt', 'getter_params', 't_def', 'ui_hint')
    s_datatype_traitclass_map = {}
    s_ui_hint = None
    s_fmt = None

    @staticmethod
    def register_by_datatype(trait_class, data_type):
        assert inspect.isclass(trait_class) and issubclass(trait_class, Trait), 'trait class must be a subclass of Trait'
        found = Trait.s_datatype_traitclass_map.get(data_type)
        assert not found, f'data_type {data_type} for {trait_class} is already registered for {found}'
        Trait.s_datatype_traitclass_map[data_type] = trait_class

    s_baseclass_traitclass_map = {}

    @staticmethod
    def register_by_baseclass(trait_class, base_class):
        assert inspect.isclass(trait_class) and issubclass(trait_class, Trait), 'trait class must be a subclass of Trait'
        assert inspect.isclass(base_class), 'base class must be a class'
        found = Trait.s_baseclass_traitclass_map.get(base_class)
        assert not found, f'base_class {base_class} for {trait_class} is already registered for {found}'
        Trait.s_baseclass_traitclass_map[base_class] = trait_class

    @staticmethod
    def real_trait_class(data_type):
        real_trait_class = Trait.s_datatype_traitclass_map.get(data_type)
        if real_trait_class:
            return real_trait_class

        tmap = Trait.s_baseclass_traitclass_map
        base_class: type
        for base_class in reversed(tmap):
            if issubclass(data_type, base_class):
                return tmap[base_class]

        return generic_trait

    def __init_subclass__(cls, data_type: type = None, register: bool = True, base_class: type = False):
        cls.s_baseclass = base_class
        if register:
            assert data_type and inspect.isclass(data_type), f'{cls} - data_type is not valid'
            if base_class:
                Trait.register_by_baseclass(cls, data_type)
            else:
                Trait.register_by_datatype(cls, data_type)

        assert cls.s_ui_hint, f'{cls} must define s_ui_hint'

    def __init__(self, t_def: TraitDefinition, btrait: BTrait = None):
        if btrait is None:
            super().__init__()
        else:
            super().__init__(btrait)

        self.t_def = t_def
        self.getter_params = ()

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if not self.getter_params:
            return instance.get_value(self)

        return functools.partial(instance.get_value, self)

    def __set__(self, instance, value):
        if not self.getter_params:
            instance.set_value(self, value).throw()

        else:
            if not isinstance(value, trait_value):
                raise TypeError(f'May not set a value to {instance.__class__.__name__}.{self.name} as it requires params')

            instance.set_value(self, value.value, *value.args).throw()

    # def __deepcopy__(self, memodict={}):
    #    return Trait(self.t_def.copy(), btrait = self)

    @staticmethod
    def create(trait_name: str, t_def: TraitDefinition) -> Trait:
        dt = t_def.data_type
        if isinstance(dt, GenericAlias):
            dt = get_origin(dt)  # get original type, e.g. `list` from `list[int]`
            # TODO: could be useful to also keep get_args(dt) for extra checking?
        trait_class = Trait.real_trait_class(dt)
        trait = trait_class(t_def)
        trait.set_name(trait_name)
        trait.data_type = dt
        trait.flags = t_def.flags.value()
        trait.default = t_def.default
        trait.fmt = t_def.fmt or trait.s_fmt

        trait.create_proc()

        trait.post_ctor()
        ui_hint: Ui = copy.deepcopy(t_def.ui_hint)
        ui_hint.adjust(trait)
        trait.ui_hint = ui_hint
        return trait

    @staticmethod
    def method_defs(trait_name: str) -> dict:
        return {
            f'{trait_name}_{(method_suffix := method_key.lower())}': (method_suffix, method_def)
            for method_key, method_def in TRAIT_METHOD.s_dir.items()
        }

    def set_trait_funcs(self, traitable_cls, rc):
        for method_name, (method_suffix, method_def) in Trait.method_defs(self.name).items():
            method = getattr(traitable_cls, method_name, None)
            if method and method_suffix == 'get' and self.t_def.default is not XNone:  # -- getter and default are defined - figure out which to use
                for cls in traitable_cls.__mro__:
                    cls_vars = vars(cls)
                    if method_name in cls_vars:  # -- found method on cls - use method, unless
                        if isinstance(cls_vars.get(self.name), TraitDefinition):  # -- default is on same cls then - error
                            rc.add_error(
                                f'Ambiguous definition for {method_name} on {cls} - both {self.name}.default and {traitable_cls}.{method_name} are defined.'
                            )
                    elif isinstance(cls_vars.get(self.name), TraitDefinition):  # -- default found on cls - use default
                        method = None  # use default
                    else:
                        continue
                    break

            f = method_def.value(self, method, method_suffix, rc)
            if f:
                set_f = getattr(self, f'set_f_{method_suffix}')
                set_f(f, bool(method))

        if self.getter_params:
            self.set_getter_has_args()

    def create_f_get(self, f, attr_name: str, rc: RC):
        if not f:  # -- no custom getter, just the default value
            f = lambda traitable: self.default_value()
            f.__name__ = 'default_value'
            params = ()

        else:
            # TODO: if default is defined in a subclass relative to where the getter is defined, override the getter?

            sig = inspect.signature(f)
            params = []
            param: Parameter
            for pname, param in sig.parameters.items():
                if pname != 'self':
                    pkind = param.kind
                    if pkind != Parameter.POSITIONAL_OR_KEYWORD:
                        rc.add_error(f'{f.__name__} - {pname} is not a positional parameter')
                    else:
                        params.append(param)
            params = tuple(params)

        self.getter_params = params
        return f

    def create_f_set(self, f, attr_name: str, rc: RC):
        if not f:
            return None

        # -- custom setter
        resolved_hints = get_type_hints(f, sys.modules[f.__module__].__dict__ if f.__module__ in sys.modules else {}, f.__class__.__dict__)
        assert resolved_hints.get('return') is RC, f'{f.__name__} - setter must return RC'

        sig = inspect.signature(f)
        params = tuple(sig.parameters.values())
        n = len(params)
        if n < 3:
            rc.add_error(f'{f.__name__} - setter must have at least 3 parameters: self, trait, value')

        getter_params = self.getter_params
        if getter_params:
            if getter_params != tuple(params[3:]):
                rc.add_error(f'{f.__name__} - setter must have same params as the getter: {getter_params}')

        return f

    def create_f_common_trait_with_value(self, f, attr_name: str, rc: RC):
        cls = self.__class__
        # TODO: check f's signature
        if not f:
            common_f = getattr(cls, attr_name, None)
            if common_f:
                f = lambda obj_or_cls, trait, value: common_f(trait, value)
                f.__name__ = f'{cls.__name__}.{common_f.__name__}'

        return f

    def create_f_common_trait_with_value_static(self, f, attr_name: str, rc: RC):
        cls = self.__class__
        if f:
            assert isinstance(f.__self__, type), f'{f.__name__} must be declared as @classmethod'
        else:
            f = getattr(cls, attr_name, None)
        return self.create_f_common_trait_with_value(f, attr_name, rc)

    def create_f_choices(self, f, attr_name: str, rc: RC):
        cls = self.__class__
        if not f:
            choices_f = getattr(cls, attr_name, None)
            if choices_f:
                f = lambda obj, trait: choices_f(trait)
                f.__name__ = f'{cls.__name__}.{choices_f.__name__}'

        return f

    def create_f_plain(self, f, attr_name: str, rc: RC):
        return f

    # =======================================================================================================================
    #   Formatting
    # =======================================================================================================================
    # fmt: off
    s_locales = {
        'Windows':      'USA',
        'Linux':        'en_US',
    }
    # fmt: on

    def locale_change(self, old_value, value):
        if value:
            return value

        try:
            return locale.setlocale(locale.LC_NUMERIC, self.__class__.s_locales.get(platform.system(), 'en_US'))
        except Exception:
            return None

    def _format(self, fmt: str) -> str:
        if not fmt:
            fmt = ':'
        else:
            c = fmt[0]
            if c != '!' and c != ':':
                fmt = ':' + fmt

        return f'{{{fmt}}}'

    def use_format_str(self, fmt: str, value) -> str:
        if isinstance(value, str) and not fmt:
            return value
        return self._format(fmt).format(value)

    # ===================================================================================================================
    #   Trait Interface
    # ===================================================================================================================

    def post_ctor(self): ...

    def check_integrity(self, cls, rc: RC):
        pass

    def default_value(self):
        return self.default

    def same_values(self, value1, value2) -> bool:
        raise NotImplementedError

    def from_any(self, value):
        if isinstance(value, self.data_type):
            return value

        if isinstance(value, str):
            return self.from_str(value)

        return self.from_any_xstr(value)

    def from_str(self, s: str):
        lit = ast.literal_eval(s)
        return self.from_any_xstr(lit)

    def from_any_xstr(self, value):
        raise NotImplementedError

    def to_str(self, v) -> str:
        return self.use_format_str(self.fmt, v)

    def is_acceptable_type(self, data_type: type) -> bool:
        return data_type is self.data_type

    def serialize(self, value):
        raise NotImplementedError

    def deserialize(self, value) -> RC:
        raise NotImplementedError

    def to_id(self, value) -> str:
        raise NotImplementedError

    def choices(self):
        return XNone

    # TODO: unify XNone/None conversions with object serialization/deserializatoin in c++
    # TODO: call these from c++ directly in place of f_serialize/f_deserialize?
    def serialize_value(self, value, replace_xnone=False):
        return None if replace_xnone and value is XNone else self.f_serialize(self, value)

    def deserialize_value(self, value, replace_none=False):
        return XNone if replace_none and value is None else self.f_deserialize(self, value)

    # ===================================================================================================================


# ---- Methods Associated with a trait
# fmt: off
class TRAIT_METHOD(NamedConstant):
    GET                 = Trait.create_f_get
    SET                 = Trait.create_f_set
    VERIFY              = Trait.create_f_plain
    FROM_STR            = Trait.create_f_common_trait_with_value
    FROM_ANY_XSTR       = Trait.create_f_common_trait_with_value
    IS_ACCEPTABLE_TYPE  = Trait.create_f_common_trait_with_value
    TO_STR              = Trait.create_f_common_trait_with_value
    SERIALIZE           = Trait.create_f_common_trait_with_value_static
    DESERIALIZE         = Trait.create_f_common_trait_with_value_static
    TO_ID               = Trait.create_f_common_trait_with_value
    CHOICES             = Trait.create_f_choices
    STYLE_SHEET         = Trait.create_f_plain
# fmt: on


class generic_trait(Trait, register=False):
    s_ui_hint = Ui.NONE

    def post_ctor(self):
        assert not self.flags_on(T.ID), f'generic trait {self.name} may not be an ID trait'
        assert self.flags_on(T.RUNTIME), f'generic trait {self.name} must be a RUNTIME trait'

    def is_acceptable_type(self, data_type: type) -> bool:
        return issubclass(data_type, self.data_type)

    def same_values(self, value1, value2) -> bool:
        return value1 is value2


class trait_value:
    def __init__(self, value, *args):
        self.value = value
        self.args = args

    def __call__(self, *args, **kwargs): ...


class BoundTrait:
    def __init__(self, obj, trait: Trait):
        self.obj = obj
        self.trait = trait
        # self.args = ()

    def __getattr__(self, attr_name):
        trait_attr = getattr(self.trait, attr_name, None)
        if trait_attr:
            # if callable(trait_attr):
            return trait_attr

        return lambda: getattr(self.obj, attr_name)(self.trait)

    def __call__(self):
        return self.trait


class ClassTrait(NamedCallable):
    def __init__(self, traitable_class, trait: Trait):
        value = lambda traitable: traitable.get_value(trait)
        super().__init__(trait.name, trait.name, value)
        self.cls = traitable_class
        self.trait = trait

    def to_str(self) -> str:
        return f'{self.cls}.{self.trait.name}'

    def serialize(self, embed: bool):
        return [
            PackageRefactoring.find_class_id(self.cls),
            self.trait.name
        ]

    @classmethod
    def deserialize(cls, serialized_data: list) -> ClassTrait:
        class_id, trait_name = serialized_data
        traitable_class = PackageRefactoring.find_class(class_id)
        trait = traitable_class.trait(trait_name, throw = True)
        return ClassTrait(traitable_class, trait)



import os
import time
from datetime import datetime
import psutil
from typing import Any
import multiprocessing as mp
import socket

from py10x_kernel import OsUser
from core_10x.traitable import Traitable, T, RT
from core_10x.xdate_time import XDateTime
from core_10x.environment_variables import EnvVars
from core_10x.resource import NULL_RESOURCE
from core_10x.ts_store import TsStore

class PerfTimer:
    def __enter__(self):
        self.start = time.perf_counter_ns()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.perf_counter_ns()
        self.elapsed = self.end - self.start

class LogMessage(Traitable, custom_collection = True, keep_history = False):
    ns: int             = T(T.ID)
    level: int          = T()
    mem_pc: float       = T()
    num_threads: int    = T()
    payload: Any        = T()

def logger_process(
        queue: mp.Queue,
        coll_name: str,
        do_print: bool
):
    uri = EnvVars.log_ts_store_uri
    log_store = TsStore.instance_from_uri(uri) if uri else NULL_RESOURCE

    data = queue.get()
    if data is None:
        return

    with log_store:
        first_msg = LogMessage(_replace = True, _collection_name = coll_name, **data)
        #start: datetime = first_msg.payload
        t1 = first_msg.ns
        first_msg.save()

    while True:
        data = queue.get()
        if data is None:
            break

        with log_store:
            msg = LogMessage(_replace = True, _collection_name = coll_name, **data)
            msg.save()

        if do_print:
            t2 = msg.ns
            dt = t2 - t1
            t1 = t2
            print(f'{(dt) / 1.e9}: {msg.payload}')

class Logger:
    def __init__(
        self,
        app_name: str,
        log_level: int,
        started_at: datetime = None,
        do_print = True
    ):
        self.app_name = app_name
        self.log_level = log_level
        if started_at is None:
            started_at = datetime.utcnow()
        self.started_at = started_at
        pid = os.getpid()
        full_name = f'{app_name}/{OsUser.me.name()}/{socket.gethostname()}/{XDateTime.datetime_to_str(started_at, False)}/{pid}'
        self.ps = psutil.Process(pid)

        self.queue = mp.Queue()
        self.proc = mp.Process(target = logger_process, args = (self.queue, full_name, do_print))
        self.proc.daemon = True
        self.proc.start()

    def log(self, data: dict):
        self.queue.put(data)

    def shutdown(self):
        self.queue.put(None)
        self.proc.join()

LOGGER: Logger = None
class LOG:
    def _log(log_level: int, payload: Any):
        assert LOGGER, 'You must have called LOG.begin(app_name, log_level)'
        if log_level > LOGGER.log_level:
            return

        if payload is not None:
            data = dict(
                ns          = time.perf_counter_ns(),
                level       = log_level,
                mem_pc      = LOGGER.ps.memory_percent(),
                num_threads = LOGGER.ps.num_threads(),
                payload     = payload
            )
        else:
            data = None

        LOGGER.log(data)

    @staticmethod
    def begin(app_name: str, log_level):
        global LOGGER
        if LOGGER is None:
            LOGGER = Logger(app_name, log_level.value)
            LOG.BRIEF(datetime.utcnow())

    @staticmethod
    def end():
        LOG._log(0, None)

    def __new__(cls, payload: Any): cls._log(0, payload)

    def BRIEF(payload: Any):    LOG._log(0, payload)
    BRIEF.value = 0
    def MEDIUM(payload: Any):   LOG._log(1, payload)
    MEDIUM.value = 1
    def DETAILED(payload: Any): LOG._log(2, payload)
    DETAILED.value = 2
    def VERBOSE(payload: Any):  LOG._log(3, payload)
    VERBOSE.value = 3

class LogReader:
    ...




from __future__ import annotations

from collections import deque
from datetime import date, timedelta

from core_10x.global_cache import cache
from core_10x.traitable import RT, T, Traitable


class CalendarNameParser:
    """
    calendar_name           := just_name | operation_repr_list
    operation_repr_list     := operation_repr | operation_repr OP_CHAR operation_repr_list
    operation_repr          := op num_args | name_list op num_args
    name_list               := just_name MORE_CHAR just_name MORE_CHAR | just_name MORE_CHAR name_list
    op                      := OR_CHAR | AND_CHAR
    """

    # fmt: off
    MORE_CHAR   = ','
    OR_CHAR     = '|'
    AND_CHAR    = '&'
    OP_CHAR     = '\n'

    s_ops = {
        OR_CHAR:    set.update,
        AND_CHAR:   set.intersection_update
    }
    # fmt: on

    @classmethod
    def combo_name(cls, name: str) -> bool:
        return any(sym in name for sym in cls.s_ops.keys())

    @classmethod
    def operation_repr(cls, calendar_cls, op_char: str, *calendars) -> str:
        assert len(calendars) > 1, 'there must be at least 2 calendars'
        assert op_char in cls.s_ops, f'Unknown op char = {op_char}'

        cal_names = []
        for cal_or_name in calendars:
            if isinstance(cal_or_name, calendar_cls):
                cname = cal_or_name.name
            elif isinstance(cal_or_name, str):
                cal = calendar_cls.existing_instance(name=cal_or_name)
                assert cal, f"Unknown calendar '{cal_or_name}'"
                cname = cal_or_name
            else:
                raise NameError(f"Invalid calendar/name '{cal_or_name}'")

            cal_names.append(cname)

        return f'{cls.MORE_CHAR.join(sorted(cal_names))}{cls.MORE_CHAR}{op_char}{len(cal_names)}{cls.OP_CHAR}'

    def __init__(self):
        self.stack = deque()
        self.non_working_days = set()

    @classmethod
    def parse(cls, calendar_cls, name: str) -> set:
        parser = cls()
        non_working_days = parser.non_working_days
        stack = parser.stack

        operation_repr_list = name.split(cls.OP_CHAR)
        if len(operation_repr_list) == 1:  # -- regular (stored) calendar
            cal = calendar_cls.existing_instance(name=name)
            return cal._non_working_days

        for operation_repr in operation_repr_list:
            if not operation_repr:
                continue

            name_list_op_num_args = operation_repr.split(cls.MORE_CHAR)
            if len(name_list_op_num_args) > 1:  # -- name list followed by op with num_args
                for cname in name_list_op_num_args[:-1]:
                    if cname:
                        cal = calendar_cls.existing_instance(name=cname)
                        assert cal, f"Unknown calendar '{cname}"
                        stack.append(cal._non_working_days)

            op_with_num_args = name_list_op_num_args[-1]
            op_char = op_with_num_args[0]
            op = cls.s_ops.get(op_char)
            assert op, f'Unknown op char {op_char}'
            try:
                num_args = int(op_with_num_args[1:])
            except Exception as e:
                raise RuntimeError(f'Invalid num_args = {op_with_num_args[1:]}') from e

            for _ in range(num_args):
                _non_working_days = stack.pop()
                assert isinstance(_non_working_days, set)

                if not non_working_days:
                    non_working_days.update(_non_working_days)
                else:
                    op(non_working_days, _non_working_days)

            stack.append(non_working_days)  # -- push set of non_working_days of an intermediate calendar

        return non_working_days


class CalendarAdjustment(Traitable):
    name: str           = T(T.ID)
    add_days: list      = T()
    remove_days: list   = T()


#-- TODO: _default_cache = True!
class Calendar(Traitable):
    name: str               = T(T.ID)
    adjusted_for: str       = T(T.ID,   default = '')   // 'Name of a specific adjustment to this calendar, if any'
    description: str        = T()                       // 'Calendar Description'
    non_working_days: list  = T()                       // 'Non-Working Days'

    _non_working_days: set  = RT()

    @staticmethod
    @cache
    def none():
        return Calendar(_replace = True, name = '__none__', non_working_days = [], description = 'Empty Calendar')

    @classmethod
    def AND(cls, *calendars) -> Calendar:  # noqa: N802
        if not calendars:
            return None

        name = CalendarNameParser.operation_repr(cls, CalendarNameParser.AND_CHAR, *calendars)
        return cls(name=name)

    intersection = AND

    @classmethod
    def OR(cls, *calendars) -> Calendar:  # noqa: N802
        if not calendars:
            return None

        name = CalendarNameParser.operation_repr(cls, CalendarNameParser.OR_CHAR, *calendars)
        return cls(name=name)

    @classmethod
    def union(cls, *calendars) -> Calendar:
        if len(calendars) > 1:
            return cls.OR(*calendars)
        assert calendars, 'At least one calendar is required for union()'
        return cls(name=calendars[0])

    def non_working_days_get(self) -> list:
        non_working_days = CalendarNameParser.parse(self.__class__, self.name)
        adjusted_for = self.adjusted_for
        if adjusted_for:
            ca = CalendarAdjustment.existing_instance_by_id(_id_value=adjusted_for, _throw=False)
            if ca:
                self.add_days(non_working_days, *ca.add_days)
                self.remove_days(non_working_days, *ca.remove_days)

        return sorted(non_working_days) if non_working_days else []

    def _non_working_days_get(self) -> set:
        return set(self.non_working_days)

    @classmethod
    def add_days(cls, days: set, *days_to_add) -> bool:
        if not days_to_add:
            return False

        all_dates = all(type(hd) is date for hd in days_to_add)
        if not all_dates:
            raise TypeError('Every day to add must be a date')

        ndays = len(days)
        days.update(days_to_add)
        return len(days) > ndays

    @classmethod
    def remove_days(cls, days: set, *days_to_remove) -> bool:
        if not days_to_remove:
            return False

        all_dates = all(type(hd) is date for hd in days_to_remove)
        if not all_dates:
            raise TypeError('Every day to remove must be a date')

        ndays = len(days)
        days.difference_update(days_to_remove)
        return len(days) < ndays

    def add_non_working_days(self, *days_to_add):
        non_working_days = set(self.non_working_days)
        if self.add_days(non_working_days, *days_to_add):
            self.non_working_days = list(non_working_days)

    def remove_non_working_days(self, *days_to_remove):
        non_working_days = set(self.non_working_days)
        if self.remove_days(non_working_days, *days_to_remove):
            self.non_working_days = list(non_working_days)

    def is_bizday(self, d: date) -> bool:
        return d not in self._non_working_days

    def next_bizday(self, d: date) -> date:
        dt = timedelta(days=1)
        d += dt
        while not self.is_bizday(d):
            d += dt

        return d

    def prev_bizday(self, d: date) -> date:
        dt = timedelta(days=-1)
        d += dt
        while not self.is_bizday(d):
            d += dt

        return d  # -- TODO: if the cal has ALL days as holidays... :-)

    def advance_bizdays(self, d: date, biz_days: int) -> date:
        if not biz_days:
            return d

        if biz_days > 0:
            for _ in range(biz_days):
                d = self.next_bizday(d)
        else:
            for _ in range(-biz_days):
                d = self.prev_bizday(d)

        return d



from __future__ import annotations

from functools import partial

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget


class TreeItem(Widget, i.TreeItem):
    __slots__ = ('handlers',)
    s_component_class = rio_components.RioTreeItem
    s_pass_children_in_kwargs = True

    def __init__(self, parent: TreeWidget | TreeItem, *args, **kwargs):
        super().__init__(*args, **kwargs)
        parent[self.s_children_attr] = parent[self.s_children_attr] + [self]
        self.handlers = parent.handlers
        for name, callback in self.handlers.items():
            self[name.replace('_item_', '_')] = partial(callback, self)

    def set_expanded(self, expanded: bool):
        self['is_expanded'] = expanded

    # noinspection PyMethodOverriding
    def set_text(self, col: int, text: str):
        super().set_text(text)

    # noinspection PyMethodOverriding
    def set_tool_tip(self, col: int, text: str):
        super().set_tool_tip(text)


class TreeWidget(Widget, i.TreeWidget):
    __slots__ = ('handlers',)
    s_component_class = rio_components.RioTreeView
    s_default_kwargs = dict(selection_mode='single')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handlers = {}

    def set_column_count(self, col_count: int):
        """Set the number of columns in the tree widget."""
        assert col_count in [1, 2], 'col_count must be 1 or 2'
        self['col_count'] = col_count

    def set_header_labels(self, labels: list):
        """Set the header labels for each column."""
        self['header_labels'] = labels

    def top_level_item_count(self) -> int:
        """Return the number of top-level items."""
        return len(self.get_children())

    def top_level_item(self, i: int) -> TreeItem:
        """Return the top-level item at index i."""
        return self.get_children()[i]

    def resize_column_to_contents(self, col: int):
        """Adjust the width of the specified column (placeholder)."""
        pass

    def item_expanded_connect(self, bound_method):
        self.handlers['on_item_expand'] = self.callback(bound_method)

    def item_clicked_connect(self, bound_method):
        # self.handlers['on_item_double_press'] = bound_method
        self.handlers['on_item_press'] = self.callback(bound_method)

    def item_pressed_connect(self, bound_method):
        self.handlers['on_item_press'] = bound_method

    def item_changed_connect(self, bound_method):
        raise NotImplementedError

    def edit_item(self, item: TreeItem, col: int):
        """Start editing the specified item in the given column (placeholder)."""
        # self.component.start_editing(item, col)
        raise NotImplementedError

    def open_persistent_editor(self, item: TreeItem, col: int):
        """Open a persistent editor for the specified item and column (placeholder)."""
        # self.component.open_persistent_editor(item, col)
        raise NotImplementedError

    def add_top_level_item(self, item: TreeItem):
        """Add a top-level item to the tree (helper method)."""
        self.get_children().append(item)



import asyncio
from datetime import date

import pytest
import rio.testing.browser_client
from core_10x.code_samples.person import Person
from core_10x.exec_control import CACHE_ONLY, INTERACTIVE
from ui_10x.collection_editor import Collection, CollectionEditor
from ui_10x.rio.component_builder import DynamicComponent


@pytest.fixture(autouse=True)
def mock_db_ops(monkeypatch):
    with CACHE_ONLY():
        sasha = Person(first_name='Sasha', last_name='Davidovich', weight_lbs=150, _replace=True)
        ilya = Person(first_name='Ilya', last_name='Pevzner', weight_lbs=200, dob=date(1971, 7, 3), _replace=True)
    monkeypatch.setattr(Person, 'load_ids', lambda: [sasha.id(), ilya.id()])
    monkeypatch.setattr(Person, 'load_data', lambda id: {sasha.id(): sasha, ilya.id(): ilya}[id])
    yield


@pytest.fixture(autouse=True)
def interactive_mode():
    interactive = INTERACTIVE()
    interactive.begin_using()
    yield
    interactive.end_using()


async def test_collection_editor() -> None:
    coll = Collection(cls=Person)
    ce = CollectionEditor(coll=coll)
    widget = ce.main_widget()

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        assert await test_client.execute_js('document.querySelector(".rio-selectable-item").innerText') == 'Ilya|Pevzner'
        await test_client.execute_js('document.querySelector(".rio-selectable-item").click()')
        await asyncio.sleep(0.5)

        assert await test_client.execute_js('document.querySelectorAll("input")[8].value') == 'LB'
        assert await test_client.execute_js('document.querySelectorAll("input")[7].value') == '200.00'

        button_id = next(b._id_ for b in test_client.get_components(rio.Button) if b.content == 'edit')
        await test_client.execute_js(f'''document.querySelector('[dbg-id="{button_id}"]').querySelector('rio-pressable-element').click()''')
        await asyncio.sleep(0.5)

        assert await test_client.execute_js('document.querySelectorAll("input")[16].value') == 'LB'
        assert await test_client.execute_js('document.querySelectorAll("input")[15].value') == '200.00'

        button_id = next(b._id_ for b in reversed(tuple(test_client.get_components(rio.Button))) if b.icon == 'material/arrow_downward')
        await test_client.execute_js(f'''document.querySelector('[dbg-id="{button_id}"]').querySelector('rio-pressable-element').click()''')
        await asyncio.sleep(0.5)

        print([li.text for li in test_client.get_components(rio.SimpleListItem)])
        list_item_id = next(li._id_ for li in test_client.get_components(rio.SimpleListItem) if li.text == 'G')
        await test_client.execute_js(f'''document.querySelector('[dbg-id="{list_item_id}"]').querySelector('.rio-selectable-item').click()''')

        assert await test_client.execute_js('document.querySelectorAll("input")[16].value') == 'G'
        assert await test_client.execute_js('document.querySelectorAll("input")[15].value') == '90,702.95'

        button_id = next(b._id_ for b in test_client.get_components(rio.Button) if b.content == 'Ok')
        await test_client.execute_js(f'''document.querySelector('[dbg-id="{button_id}"]').querySelector('rio-pressable-element').click()''')
        await asyncio.sleep(0.5)

        assert await test_client.execute_js('document.querySelectorAll("input")[8].value') == 'G'
        assert await test_client.execute_js('document.querySelectorAll("input")[7].value') == '90,702.95'



#!/usr/bin/env python3
"""
In-memory TestStore for testing TraitableHistory and other storage functionality.

This provides a lightweight, in-memory implementation of the TsStore interface
that can be used for unit tests without external dependencies like MongoDB.
"""

from __future__ import annotations

import copy
from datetime import datetime
import re
from typing import TYPE_CHECKING

from py10x_kernel import BTraitable, BTraitableProcessor

from core_10x.nucleus import Nucleus
from core_10x.ts_store import TsCollection, TsDuplicateKeyError, TsStore

if TYPE_CHECKING:
    from collections.abc import Iterable

    from core_10x.trait_filter import f


class TestCollection(TsCollection):
    """In-memory collection implementation for testing."""

    s_id_tag = '_id'

    def __init__(self, store: TestStore, collection_name: str):
        self.store = store
        self._collection_name = collection_name
        self._documents = {}  # id -> document
        self._indexes = {}  # index_name -> index_info

    def collection_name(self) -> str:
        return self._collection_name

    def _effective_documents(self) -> dict:
        """Documents visible in the current transaction (committed + pending, minus pending deletes)."""
        out = dict(self._documents)
        tx = self.store.current_transaction()
        if tx is not None:
            pending = tx.pending_saves.get(self._collection_name, {})
            deletes = tx.pending_deletes.get(self._collection_name, set())
            out.update(pending)
            for id_value in deletes:
                out.pop(id_value, None)
        return out

    def id_exists(self, id_value: str) -> bool:
        """Check if a document with the given ID exists."""
        return id_value in self._effective_documents()

    def _eval(self, doc, query):
        bclass = query.traitable_class
        try:
            return query.eval(doc)
        except AttributeError:
            # TODO: fix - need to deserialize as dictionary-based eval doesn't support nested objects
            if not bclass:
                raise  # likely and ID query from load(ID)
            coll = self._collection_name if bclass.is_custom_collection() else None
            with BTraitableProcessor.create_root():
                return query.eval(BTraitable.deserialize_object(bclass, coll, copy.copy(doc)))

    def find(self, query: f = None, _at_most: int = 0, _order: dict = None) -> Iterable:
        """Find documents matching the query."""
        documents = list(self._effective_documents().values())

        # Apply query filter if provided
        if query:
            documents = [doc for doc in documents if self._eval(doc, query)]

        # Apply ordering if provided
        if _order:
            # Sort by all fields at once, with the last field taking precedence
            # We need to reverse the order of fields so the last one is applied first
            sort_fields = list(_order.items())
            sort_fields.reverse()  # Last field first

            for field, direction in sort_fields:
                documents.sort(key=lambda x: x.get(field, ''), reverse=(direction == -1))

        # Apply limit if specified
        if _at_most > 0:
            documents = documents[:_at_most]

        return (dict(doc) for doc in documents)

    def count(self, query: f = None) -> int:
        """Count documents matching the query."""
        return len(list(self.find(query)))

    def save_new(self, serialized_traitable: dict, overwrite: bool = False) -> int:
        """Save a new document."""

        # Handle MongoDB-style operations
        if '$set' in serialized_traitable:
            # This is a MongoDB-style update operation

            # Extract the data from $set
            data = serialized_traitable['$set'].copy()

            # Handle $currentDate operations
            if '$currentDate' in serialized_traitable:
                current_date_fields = serialized_traitable['$currentDate']
                for field in current_date_fields:
                    if current_date_fields[field] is True:
                        data[field] = datetime.utcnow()

            serialized_traitable = data

        serialized_traitable[Nucleus.REVISION_TAG()] = 1

        id_tag = self.s_id_tag
        id_value = serialized_traitable.get(id_tag)

        if not id_value:
            return 0

        tx = self.store.current_transaction()
        if tx is not None:
            effective = self._effective_documents()
            if id_value in effective and not overwrite:
                raise TsDuplicateKeyError(self.collection_name(), {id_tag: id_value})
            if self._collection_name not in tx.pending_saves:
                tx.pending_saves[self._collection_name] = {}
            tx.pending_deletes.get(self._collection_name, set()).discard(id_value)
            tx.pending_saves[self._collection_name][id_value] = dict(serialized_traitable)
            return 1

        if id_value in self._documents and not overwrite:
            raise TsDuplicateKeyError(self.collection_name(), {id_tag: id_value})

        self._documents[id_value] = serialized_traitable
        return 1

    def save(self, serialized_traitable: dict) -> int:
        """Save or update a document."""
        rev_tag = Nucleus.REVISION_TAG()
        revision = serialized_traitable[rev_tag]
        if revision == 0:
            return self.save_new(serialized_traitable)

        undef_variable = next((k[1:] for k in serialized_traitable if k.startswith('$')), None)
        if undef_variable:
            raise RuntimeError(f'Use of undefined variable: {undef_variable}')
        doc_id = serialized_traitable.get(self.s_id_tag)
        assert doc_id

        effective = self._effective_documents()
        existing_doc = effective.get(doc_id)
        existing_revision = existing_doc[rev_tag] if existing_doc else 0
        assert revision == existing_revision

        if existing_doc == serialized_traitable:
            return existing_revision

        revision += 1
        new_doc = serialized_traitable | {rev_tag: revision}

        tx = self.store.current_transaction()
        if tx is not None:
            if self._collection_name not in tx.pending_saves:
                tx.pending_saves[self._collection_name] = {}
            tx.pending_deletes.get(self._collection_name, set()).discard(doc_id)
            tx.pending_saves[self._collection_name][doc_id] = new_doc
            return revision

        self._documents[doc_id] = new_doc
        return revision

    def delete(self, id_value: str) -> bool:
        """Delete a document by ID."""
        tx = self.store.current_transaction()
        if tx is not None:
            effective = self._effective_documents()
            if id_value not in effective:
                return False
            if self._collection_name not in tx.pending_deletes:
                tx.pending_deletes[self._collection_name] = set()
            tx.pending_deletes[self._collection_name].add(id_value)
            tx.pending_saves.get(self._collection_name, {}).pop(id_value, None)
            return True
        if id_value in self._documents:
            del self._documents[id_value]
            return True
        return False

    def create_index(self, name: str, trait_name: str | list[tuple[str, int]], **index_args) -> str:
        """Create an index (in-memory implementation just stores the index info)."""
        self._indexes[name] = {'trait_name': trait_name, 'args': index_args}
        return name

    def max(self, trait_name: str, filter: f = None) -> dict:
        """Find the document with the maximum value for the given trait."""
        documents = list(self.find(filter))
        if not documents:
            return {}

        max_doc = max(documents, key=lambda x: x.get(trait_name, ''))
        return max_doc

    def min(self, trait_name: str, filter: f = None) -> dict:
        """Find the document with the minimum value for the given trait."""
        documents = list(self.find(filter))
        if not documents:
            return {}

        min_doc = min(documents, key=lambda x: x.get(trait_name, ''))
        return min_doc



class TestStore(TsStore, resource_name='TEST_DB'):
    """In-memory store implementation for testing."""

    class Transaction(TsStore.Transaction):
        """In-memory transaction: commit() applies pending writes, abort() discards them."""

        def __init__(self, store: TestStore):
            if not (current_tx:=store.current_transaction()):
                self.pending_saves: dict[str, dict[str, dict]] = {}  # coll_name -> id -> doc
                self.pending_deletes: dict[str, set[str]] = {}  # coll_name -> set of id
            else:
                self.pending_saves = current_tx.pending_saves
                self.pending_deletes = current_tx.pending_deletes
            super().__init__(store)

        def _do_commit(self) -> None:
            if self.store.current_transaction():
               return #-- no nested transactions supported
            for coll_name, docs in self.pending_saves.items():
                if coll_name not in self.store._collections:
                    continue
                coll = self.store._collections[coll_name]
                for id_value, doc in docs.items():
                    coll._documents[id_value] = doc
            for coll_name, ids in self.pending_deletes.items():
                if coll_name not in self.store._collections:
                    continue
                coll = self.store._collections[coll_name]
                for id_value in ids:
                    coll._documents.pop(id_value, None)

        def _do_abort(self) -> None:
            pass

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._collections = {}
        self._collection_names = set()
        # Set a default username for testing
        self.username = kwargs.get('username', 'test_user')

    @classmethod
    def new_instance(cls, *args, password: str = '', **kwargs) -> TestStore:
        """Create a new TestStore instance."""
        return cls(*args, **kwargs)

    @classmethod
    def parse_uri(cls, uri: str) -> dict:
        """Parse a test store URI (e.g. testdb: or testdb://). Returns empty kwargs."""
        if not uri.startswith('testdb:'):
            raise ValueError(f'Invalid TestStore URI: {uri}')
        return {}

    def collection_names(self, regexp: str = None) -> list:
        """Get collection names, optionally filtered by regexp."""
        names = list(self._collection_names)

        if regexp:
            pattern = re.compile(regexp)
            names = [name for name in names if pattern.match(name)]

        return sorted(names)

    def collection(self, collection_name: str) -> TestCollection:
        """Get or create a collection."""
        if collection_name not in self._collections:
            self._collections[collection_name] = TestCollection(self, collection_name)
            self._collection_names.add(collection_name)

        return self._collections[collection_name]

    def delete_collection(self, collection_name: str) -> bool:
        """Delete a collection."""
        if collection_name in self._collections:
            del self._collections[collection_name]
            self._collection_names.discard(collection_name)
            return True
        return False

    @classmethod
    def is_running_with_auth(cls, host_name: str) -> tuple:
        """TestStore doesn't require authentication."""
        return True, False

    def clear(self):
        """Clear all collections (useful for test cleanup)."""
        self._collections.clear()
        self._collection_names.clear()

    def get_document_count(self, collection_name: str) -> int:
        """Get the number of documents in a collection (for testing)."""
        if collection_name in self._collections:
            return len(self._collections[collection_name]._documents)
        return 0

    def get_documents(self, collection_name: str) -> list:
        """Get all documents in a collection (for testing)."""
        if collection_name in self._collections:
            return list(self._collections[collection_name]._documents.values())
        return []

    def auth_user(self) -> str | None:
        return self.username




from datetime import date, datetime

from infra_10x.mongodb_store import MongoStore

from core_10x.code_samples.person import Person

if __name__ == '__main__':
    with MongoStore.instance(hostname='localhost', dbname='traitable_history_test'):
        Person.delete_collection()
        Person.s_history_class.delete_collection()

        # Create and save a person
        person = Person(first_name='Alyssa', last_name='Lees', dob=date(1985, 7, 5), _replace=True)
        person.save()

        ts = datetime.utcnow()

        # Update and save again
        person.dob = date(1985, 7, 6)
        person.save()

        # Check history
        history = Person.history()

        # Verify server-side fields
        for entry in history:
            print(f'  - Revision {entry["_traitable_rev"]}: {entry["first_name"]} (dob: {entry["dob"]})')
            print(f'    _who: {entry["_who"]}, _at: {entry["_at"]}')

        assert person.dob == date(1985, 7, 6)

        print(Person.restore(person.id(), timestamp=ts))
        assert person.dob == date(1985, 7, 5)

        person.reload()
        assert person.dob == date(1985, 7, 6)

        print(Person.restore(person.id(), timestamp=ts, save=True))
        assert person.dob == date(1985, 7, 5)
        person.reload()
        assert person.dob == date(1985, 7, 5)



from core_10x.trait_definition import M
from core_10x.traitable import T, Traitable
from core_10x.xnone import XNone


class X(Traitable):
    x: int = T(T.ID | T.RUNTIME)
    v: int = T(T.ID | T.RUNTIME)

    def v_get(self):
        return self.x + 1


class Y(X):
    v: int = M(flags=(None, T.ID), default=3)


class Z(Y):
    v: float = M(default=XNone)

    def v_get(self):
        return self.x + 3


if __name__ == '__main__':
    x = X(x=1)
    y = Y(x=1)
    z = Z(x=1)
    print(x, x.v)
    print(y, y.v)
    print(z, z.v)




# NOTES:
# current classes trait.default override base classes trait_get
# current classes trait_get, overrides base classes trait.default, except if there is a trait modification  in this class, in which case, it needs to reset trait.default
# an error is thrown if trait_get and trait.default are defined in the same class (or there is a trait modification in the same class where trait_get is defined that does not reset trait.default from base class)


from __future__ import annotations

import inspect
import io
import pickle
import shlex
from typing import TYPE_CHECKING

from importlib_resources import files

from core_10x.global_cache import cache

if TYPE_CHECKING:
    from collections.abc import Callable


class PyClass:
    """
    Module top-level classes ONLY
    """

    # fmt: off
    NO_NAME         = lambda cls: ''
    QUAL_NAME       = lambda cls: cls.__qualname__
    CANONICAL_NAME  = lambda cls: f'{cls.__module__}.{cls.__qualname__}'
    # fmt: on

    @staticmethod
    def name(cls, name_type: Callable[[type], str] = CANONICAL_NAME) -> str:
        try:
            return name_type(cls)
        except Exception as ex:
            raise ValueError('cls must be a valid class') from ex

    @staticmethod
    def top_level_package(cls) -> str:
        return cls.__module__.split('.', maxsplit=1)[0]

    # ===================================================================================================================
    #   Finding classes/symbols in the code base
    # ===================================================================================================================
    @staticmethod
    @cache
    def dummy_unpickler() -> pickle.Unpickler:
        file = io.BytesIO()
        return pickle.Unpickler(file)

    @staticmethod
    @cache
    def find_symbol(canonical_symbol_name: str):
        try:
            module_name, symbol_name = canonical_symbol_name.rsplit('.', maxsplit=1)
        except Exception as e:
            raise ValueError(f"Invalid canonical_symbol_name = '{canonical_symbol_name}'") from e

        try:
            return PyClass.dummy_unpickler().find_class(module_name, symbol_name)

        except Exception:
            return None

    @staticmethod
    def find(canonical_class_name: str, *parents):
        cls = PyClass.find_symbol(canonical_class_name)
        if not cls or not inspect.isclass(cls):
            return None

        subclass = all(issubclass(cls, parent) for parent in parents)
        return cls if subclass else None

    @staticmethod
    def find_by_topic_and_suffix(topic: str, suffix: str, package_name: str, module_name: str, class_name: str):
        module_name_with_topic = f'{module_name}_{topic}'
        class_name_with_suffix = f'{class_name}{suffix}'
        found = PyClass.find(f'{package_name}.{module_name_with_topic}.{class_name_with_suffix}')
        if not found:
            found = PyClass.find(f'{package_name}.{topic}.{module_name_with_topic}.{class_name_with_suffix}')

        return found

    @staticmethod
    @cache
    def find_related_class(cls: type, topic: str, class_name_suffix: str, *alternative_packages, alternative_parent_class: type = None) -> type:
        """
        Tries to find a class whose name is cls.__name__ + class_name_suffix "related" to cls by a topic (e.g., 'ui').
        By convention the module name must be the cls' module short name + _ + topic.
        First, tries the alternative_packages, if any, then cls-module-package, then the cls-module-package.topic.
        If fails, to find, tries the same logic for alternative_parent_class (cls must be a subclass of it), if any.

        For example:

        Consider class TextMessage in module abc.infra.messenger. Its custom editor class, will be looked up as follows:
            PyClass.find_related_class(TextMessage, 'ui', 'Editor', *extra_packages)
                - editor class name:            TextMessageEditor
                - editor short module name:     messenger_ui
                - search for TextMessageEditor in the following order:
                    -- [ ex_package.messenger_ui for ex_package in extra_packages ]
                    -- the module abc.infra.messenger_ui
                    -- the module abc.infra.ui.messenger_ui
        """
        parts = cls.__module__.rsplit('.', maxsplit=1)
        assert len(parts) >= 2, f'{cls} - package is missing'

        module_name = f'{parts[-1]}'
        class_name = cls.__name__

        # -- 1) look it up in alternative_packages
        for alt_package in alternative_packages:
            found = PyClass.find_by_topic_and_suffix(topic, class_name_suffix, alt_package, module_name, class_name)
            if found:
                return found

        # -- 2) look it up in the cls' package
        package_name = parts[0]
        found = PyClass.find_by_topic_and_suffix(topic, class_name_suffix, package_name, module_name, class_name)
        if found:
            return found

        # -- 3) look it up for alternative_parent_class, if any
        if alternative_parent_class:
            assert issubclass(cls, alternative_parent_class), f'{cls} is not a subclass of {alternative_parent_class}'
            found = PyClass.find_related_class(alternative_parent_class, topic, class_name_suffix)

        return found

    @staticmethod
    def derived_from(cls, *parents, exclude_parents: tuple = ()) -> bool:
        """
        :param cls: a class
        :param parents: classes the cls must be derived from
        :param exclude_parents: classes the cls must NOT be derived from
        """
        if any(issubclass(cls, parent) for parent in exclude_parents):
            return False

        return all(issubclass(cls, parent) for parent in parents)

    @staticmethod
    def parents(cls) -> tuple:
        tree = inspect.getclasstree([cls])
        try:
            return tree[-1][0][1]
        except Exception as e:
            raise AssertionError(f'Something went wrong with inheritance tree of class {PyClass.name(cls)}') from e

    @staticmethod
    def class_tree(root_class: type, *classes) -> dict:
        all_nodes = {}
        for cls in classes:
            PyClass._collect_class_nodes(cls, root_class, all_nodes)

        return all_nodes.get(root_class)

    @staticmethod
    def _collect_class_nodes(cls: type, root_class: type, all_nodes: dict):
        if not issubclass(cls, root_class):
            return

        node = all_nodes.get(cls)
        if node is not None:
            return

        node = {}
        all_nodes[cls] = node
        parents = PyClass.parents(cls)
        for p in parents:
            PyClass._collect_class_nodes(p, root_class, all_nodes)
            p_node = all_nodes.get(p)
            if p_node is not None:
                p_node[cls] = node

    @staticmethod
    def inheritance_paths(root_class: type, child_class: type) -> list:
        tree = PyClass.class_tree(root_class, child_class)
        return PyClass._inheritance_paths(tree)

    @staticmethod
    def _inheritance_paths(tree: dict) -> list:
        res = []
        for cls, class_entry in tree.items():
            if not class_entry:
                res.append([cls])
                continue

            cls_paths = PyClass._inheritance_paths(class_entry)
            for path in cls_paths:
                x_path = [cls]
                x_path.extend(path)
                res.append(x_path)

        return res

    @staticmethod
    def full_name_space(cls) -> dict:
        full_ns = {}
        classes = *PyClass.parents(cls), cls
        for c in classes:
            ns = dict(vars(inspect.getmodule(c)))
            full_ns.update(ns)

        return full_ns

    # ---- We may need it to create local vars with particular names and values
    # locals().update( { var_name: var_value } )

    @staticmethod
    def module_class_names(package_name: str, py_file_name: str) -> list:
        try:
            res = []
            src = files(package_name).joinpath(py_file_name).read_text()
            for line in src.split('\n'):
                if line.startswith('class'):
                    class_name = shlex.shlex(line[5:]).get_token()
                    if class_name.isidentifier():
                        res.append(class_name)

            return res

        except Exception:
            return []

    @staticmethod
    def class_names_by_module(*package_names, exclude_packages=()) -> dict:
        res = {}
        for pname in package_names:
            PyClass._collect_class_names(res, pname, exclude_packages)

        return res

    @staticmethod
    def canonical_class_names(*package_names, exclude_packages=()) -> list:
        res = []
        for pname in package_names:
            PyClass._collect_class_names(res, pname, exclude_packages)

        return res

    @staticmethod
    def _collect_class_names(res, package_name: str, exclude_packages: tuple):
        if package_name in exclude_packages:
            return

        try:
            dir = files(package_name)

        except Exception as e:
            raise AssertionError(f"'{package_name}' is neither a package, nor a module") from e

        for item in dir.iterdir():
            name: str = item.name
            if name.startswith('__'):
                continue

            if item.is_dir():
                if isinstance(res, dict):
                    subres = {}
                    PyClass._collect_class_names(subres, package_name + '.' + name, exclude_packages)
                    if subres:
                        res[name] = subres

                else:
                    PyClass._collect_class_names(res, package_name + '.' + name, exclude_packages)

                continue

            if name.endswith('.py'):
                PyClass._collect_module_class_names(res, package_name, name)
                continue

    @staticmethod
    def _collect_module_class_names(res, package_name: str, py_file_name: str):
        class_names = PyClass.module_class_names(package_name, py_file_name)
        if class_names:
            module_name = py_file_name[:-3]
            if isinstance(res, dict):
                res[module_name] = class_names
            else:
                full_module_name = f'{package_name}.{module_name}'
                res.extend([f'{full_module_name}.{cname}' for cname in class_names])

    @staticmethod
    def all_classes(*package_names, exclude_packages=(), parent_classes=()) -> tuple:
        """
        Warning! This function will import every class in the packages specified.
        """
        all_class_names = PyClass.canonical_class_names(*package_names, exclude_packages=exclude_packages)
        return tuple(cls for class_name in all_class_names if (cls := PyClass.find(class_name, *parent_classes)))

    @staticmethod
    def class_name_tree(classes: tuple) -> dict:
        dir = {}
        for cls in classes:
            module_name = cls.__module__
            path = module_name.split('.')
            d = dir
            for package in path[:-1]:
                d = d.setdefault(package, {})
            module_classes = d.setdefault(path[-1], [])
            module_classes.append(cls.__qualname__)

        return dir

    @staticmethod
    def own_attribute(cls, attr_name: str) -> tuple:  # -- (exists, value)
        d = cls.__dict__
        value = d.get(attr_name, d)
        rc = value is not d
        return (rc, value if rc else None)

    # ===================================================================================================================
    #   Class mapping to deal with migrations / refactoring class canonical names
    # ===================================================================================================================


#     s_migration_map     = {}
#     s_rev_migration_map = {}
#     @staticmethod
#     def register_migration(canonical_class_name: str, new_canonical_class_name: str):
#         existing_mapping    = PyClass.s_migration_map.get(canonical_class_name)
#         reverse_mapping     = PyClass.s_rev_migration_map.get(new_canonical_class_name)
#
#         if not existing_mapping:
#             PyClass.s_migration_map[canonical_class_name] = new_canonical_class_name
#             PyClass.s_rev_migration_map[new_canonical_class_name] = canonical_class_name
#
#         else:
#             assert existing_mapping == new_canonical_class_name, f'{canonical_class_name} is already migrated to {existing_mapping}'
#             assert reverse_mapping == canonical_class_name, f'{new_canonical_class_name} is mapped to {reverse_mapping}'
#
#         module_name, class_name = canonical_class_name.rsplit('.', 1)
#         new_module_name, new_class_name = new_canonical_class_name.rsplit('.', 1)
#
#         deferred_module = sys.modules.get(module_name)
#         if not isinstance(deferred_module, DeferredModule):
#             assert not hasattr(deferred_module, class_name), f'{class_name} is already defined in {module_name}'
#             deferred_module = DeferredModule(module_name)
#
#         if class_name not in deferred_module.__dict__:
#             setattr(deferred_module, class_name, DeferredClass(module_name = new_module_name, class_name = new_class_name))
#
#     @staticmethod
#     def register_multiple_migrations(migration_map: dict):
#         for canonical_class_name, new_canonical_class_name in migration_map.items():
#             PyClass.register_migration(canonical_class_name, new_canonical_class_name)
#
#     @staticmethod
#     def all_migrations(canonical_class_name: str) -> list:
#         rev_map = PyClass.s_rev_migration_map
#         all_names = []
#         while canonical_class_name:
#             all_names.append(canonical_class_name)
#             canonical_class_name = rev_map.get(canonical_class_name)
#
#         return all_names
#
#     @staticmethod
#     @cache
#     def effective_canonical_name(cls) -> str:
#         return PyClass.all_migrations(PyClass.name(cls))[-1]
#
# class DeferredModule(type(sys)):
#     def __init__(self, module_name: str):
#         super().__init__(module_name)
#         self.deferred_count = None
#         self.module = sys.modules.get(module_name)
#
#         try:
#             self.__spec__ = importlib.util.find_spec(module_name)
#         except ModuleNotFoundError:
#             pass
#
#         sys.modules[module_name] = self
#         parent, _, name = module_name.rpartition('.')
#         assert parent, f'{module_name} - top level deferred modules are not allowed'
#
#         parent_module = sys.modules.get(parent)
#         if not parent_module:
#             top_level = parent.rpartition('.')[0]
#             parent_module = DeferredModule(parent) if top_level else importlib.import_module(parent)
#
#         self.parent = parent
#         self.name = name
#
#         setattr(parent_module, name, self)
#
#     def __getattribute__(self, item):
#         if item == '__path__':
#             parent_path = sys.modules[self.parent].__path__
#             name = self.name
#             return [ f'{p_path}/{name}' for p_path in parent_path ]
#
#         if item[0] in ( 'm', '_' ) and item[1] == '_':
#             return object.__getattribute__(self, item)
#
#         module_name = self.__name__
#
#         d_count = self.deferred_count
#         if d_count is None:
#             assert self is sys.modules[module_name], f'Module {module_name} is different from sys.modules[{module_name}]'
#             module = self.module
#             if module:
#                 getattr(module, item)   #-- import it as it's deferred
#                 self.__dict__.update(module.__dict__)
#             else:
#                 del sys.modules[module_name]
#                 spec = importlib.util.find_spec(module_name)
#                 if spec:
#                     module = importlib.import_module(module_name)
#                     self.__dict__.update(module.__dict__)
#
#                 sys.modules[module_name] = self
#                 setattr(sys.modules[self.parent], self.name, self)
#
#             d_count = sum( isinstance(value, DeferredClass) for value in self.__dict__.values() )
#
#         df_class = self.__dict__.get(item)
#         if isinstance(df_class, DeferredClass):
#             self.__dict__[item] = df_class.finalize()
#             d_count -= 1
#
#         self.deferred_count = d_count
#         return object.__getattribute__(self, item)
#
# class DeferredClass:
#     def __init__(self, module_name: str = None, class_name: str = None):
#         self.module_name = module_name
#         self.class_name  = class_name
#
#     def finalize(self) -> type:
#         module = importlib.import_module(self.module_name)
#         return getattr(module, self.class_name)



import re

from core_10x.resource import TS_STORE, ResourceRequirements


def _raise_type_error(cls, **kwargs):
    raise TypeError(f'{cls} - subclassing is not allowed')


class DataDomain:
    s_dir = {}

    s_resource_requirements = {}

    def __init_subclass__(cls, **kwargs):
        name = cls.name()
        ed = DataDomain.s_dir.get(name)
        assert ed is None, f"DataDomain '{name}' is already registered in {ed}"
        DataDomain.s_dir[name] = cls

        cls.__init_subclass__ = lambda **kwargs: _raise_type_error(cls, **kwargs)
        category: str
        rrs = cls.s_resource_requirements
        for category, resource_reqs in cls.__dict__.items():
            if isinstance(resource_reqs, ResourceRequirements):
                assert category.isupper(), f"Category '{category}' must be an uppercase name"
                resource_reqs.domain = cls
                resource_reqs.category = category
                rrs[category] = resource_reqs

        assert rrs, f"DataDomain '{cls} must define at least one category"

    @classmethod
    def name(cls: str) -> str:
        return re.sub(r'([a-z])([A-Z])', r'\1_\2', cls.__qualname__).upper()

    @classmethod
    def resource_requirement(cls, category: str, throw: bool = True) -> ResourceRequirements:
        rr = cls.s_resource_requirements.get(category)
        if rr is None and throw:
            raise ValueError(f"DataDomain {cls} - unknown category '{category}'")

        return rr


class GeneralDomain(DataDomain):
    GENERAL = TS_STORE()
    ...



from core_10x.rc import RC_TRUE

if __name__ == '__main__':
    #import os;assert os.environ.setdefault('UI_PLATFORM', 'Rio') == os.getenv('UI_PLATFORM')

    from core_10x.named_constant import NamedConstant

    from ui_10x.utils import UxDialog, UxRadioBox, ux

    class COLOR(NamedConstant, lowercase_values = True):
        BLACK   = ()
        WHITE   = ()
        GREEN   = ()
        RED     = ()
        BROWN   = ()

    app = ux.init()

    w = UxRadioBox(COLOR, 'Choose a Color', default_value = COLOR.GREEN)
    def cb():
        print('accept:', w.choice() )
        return RC_TRUE
    d = UxDialog(w, accept_callback=cb)

    rc = d.exec()

    print(w.choice(),rc)





from __future__ import annotations

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import MouseEvent, Widget


class LineEdit(Widget, i.LineEdit):
    s_default_kwargs = dict(text='')
    s_component_class = rio_components.LineEditComponent
    s_single_child = True
    s_children_attr = 'text'

    def set_text(self, text: str):
        self['text'] = text or ''

    def text(self):
        return self['text']

    def text_edited_connect(self, bound_method):
        self['on_change'] = self.callback(lambda ev: bound_method(ev.text))

    def set_password_mode(self):
        self['is_secret'] = True

    def editing_finished_connect(self, bound_method):
        self['on_lose_focus'] = self.callback(lambda ev: bound_method())

    def _make_kwargs(self, **kwargs):
        kw = super()._make_kwargs(**kwargs)
        handler = self.mouse_press_event
        if handler.__func__ != LineEdit.mouse_press_event:  # do not set unless re-implemented
            kw['on_pointer_up'] = self.callback(lambda event: handler(MouseEvent(event)))
        return kw

    def mouse_press_event(self, event: i.MouseEvent):
        pass



from __future__ import annotations

import copy
import inspect
from typing import TYPE_CHECKING

from core_10x.global_cache import cache
from core_10x.trait import T
from core_10x.traitable import Traitable
from core_10x.ui_hint import Ui, UiMod

if TYPE_CHECKING:
    from core_10x.trait import Trait


class TraitableView:
    """
    To create an instance of TraitableView you need to pass the following args:
    1) entity_or_class: a subclass of Entity or an instance of it
    2) _header_data: a dictionary with each item as follows:
        either
            <trait-name>:   <label>     (if label is empty, the trait's label will be used)
        or
            <group-label>:  <header>    (_header like dict)
    3) named_traits_to_use: (kwargs) with each item as follows:
            <trait-name>:   Ui(...)

        Such named_traits specify either traits to use (slice) or the ones to modify (modify)
    """

    @classmethod
    def _check(cls, hint_mod, trait_name: str) -> bool:
        assert isinstance(hint_mod, UiMod), f'{trait_name} = {hint_mod} - UiMod is expected'
        return True

    @classmethod
    def suitable_record(cls, trait: Trait = None, trait_dir: dict = None, trait_name: str = None, _skip_reserved: bool = True) -> Trait | None:
        if not trait:
            assert trait_dir and trait_name, 'trait is None, so both trait_dir and trait_name must be provided'
            trait = trait_dir.get(trait_name)
            if not trait:
                return None

        flags = T.HIDDEN | T.RESERVED if _skip_reserved else T.HIDDEN
        return None if trait.flags_on(flags) or trait.getter_params or trait.s_ui_hint.widget_type == Ui.WIDGET_TYPE.NONE else trait

    @classmethod
    def slice(cls, traitable_class, _header_data: dict = None, **named_ui_hint_changes) -> TraitableView:
        assert inspect.isclass(traitable_class) and issubclass(traitable_class, Traitable), f'{traitable_class} is not a Traitable'

        trait_dir = traitable_class.s_dir
        trait: Trait
        ui_hints = {
            name: hint_change.apply(trait.ui_hint) if cls._check(hint_change, trait.name) else None
            for name, hint_change in named_ui_hint_changes.items()
            if (trait := cls.suitable_record(trait_dir=trait_dir, trait_name=name))
        }
        return TraitableView(traitable_class, ui_hints, _header_data)

    @classmethod
    def modify(
        cls, traitable_class: type[Traitable], _header_data: dict = None, _skip_reserved: bool = True, **named_ui_hint_changes
    ) -> TraitableView:
        assert inspect.isclass(traitable_class) and issubclass(traitable_class, Traitable), f'{traitable_class} is not a Traitable'

        ui_hints = {
            trait_name: trait.ui_hint
            for trait_name, trait in traitable_class.s_dir.items()
            if TraitableView.suitable_record(trait=trait, _skip_reserved=_skip_reserved)
        }

        for name, hint_change in named_ui_hint_changes.items():
            cls._check(hint_change, name)
            hint = ui_hints.get(name)
            if hint is not None:
                ui_hints[name] = hint_change.apply(hint)

        return TraitableView(traitable_class, ui_hints, _header_data)

    @staticmethod
    @cache
    def default(traitable_class: type[Traitable], read_only: bool = False) -> TraitableView:
        return TraitableView.modify(traitable_class).make_read_only(read_only, clone=True)

    def make_read_only(self, flag: bool, clone=False) -> TraitableView:
        view = self if not clone else copy.deepcopy(self)

        to_set = Ui.READ_ONLY if flag else 0x0
        to_reset = 0x0 if flag else Ui.READ_ONLY
        for ui_hint in view.ui_hints.values():
            ui_hint.set_reset_flags(to_set, to_reset)
        return view

    def __init__(self, traitable_class, ui_hints: dict, header_data: dict):
        self.cls = traitable_class
        self.ui_hints = ui_hints
        self.header = self.create_header(ui_hints, header_data)

    def all_hints(self) -> dict:
        return self.ui_hints

    def ui_hint(self, trait_name: str) -> Ui:
        return self.ui_hints[trait_name]

    @staticmethod
    def _process_tree(ui_hints: dict, tree: dict) -> dict:
        header = {}
        for subtree_name, subtree in tree.items():
            if isinstance(subtree, str):  # -- must be a trait name
                label = subtree
                hint = ui_hints.get(subtree_name)
                if not hint:  # -- ignore an unknown or unused trait
                    continue
                if not label:  # -- no label overwrite
                    label = hint.label
                header[subtree_name] = label

            elif isinstance(subtree, dict):  # -- must be a real subtree
                header[subtree_name] = TraitableView._process_tree(ui_hints, subtree)

            else:
                continue  # -- ignore anything else

        return header

    @staticmethod
    def create_header(ui_hints: dict, header_data: dict) -> dict:
        if not header_data:
            return {trait_name: ui.label for trait_name, ui in ui_hints.items()}

        return TraitableView._process_tree(ui_hints, header_data)



import keyring

from core_10x.ts_store import TsStore

db1 = TsStore.instance_from_uri('mongodb://localhost:27017/mkt_data')
print(db1.collection_names())

user = 'admin'
db2 = TsStore.instance_from_uri(f'mongodb://{user}:{keyring.get_password("mongodb_local", user)}@localhost:27018/test')
print(db2.collection_names())



from core_10x.traitable import RC, RC_TRUE, RT, Traitable


class Example(Traitable):
    code: str = RT()
    limit: int = RT(0)

    def code_verify(self, t, value: str) -> RC:
        """Run on verify() or save(); not on set."""
        if value and not value.isalnum():
            return RC(False, "code must be alphanumeric")
        return RC_TRUE

    def limit_verify(self, t, value: int) -> RC:
        if value is not None and value > 100:
            return RC(False, "limit must be <= 100")
        return RC_TRUE


if __name__ == "__main__":
    # Invalid values can be set; verification fails when requested
    e = Example()
    e.code = "invalid-code"
    e.limit = 150
    assert e.code == "invalid-code"
    assert e.limit == 150

    rc = e.verify()
    assert not rc  # fails due to code_verify and limit_verify

    # Fix and verify
    e.code = "valid"
    e.limit = 50
    e.verify().throw()  # success



from datetime import datetime
import itertools
import random

from core_10x.basket import Basket, Basketable, BUCKET_SHAPE, BucketDict
from core_10x.traitable import Traitable, T, RT
from core_10x.trait_filter import f


class FinInstrument(Traitable, Basketable, bucket_shape = BUCKET_SHAPE.DICT):
    name: str       = T(T.ID)
    leaves: dict    = T()

    def price(self):
        return sum(leaf.price() * qty for leaf, qty in self.leaves.items())

    def members_qtys(self):
        return self.leaves.items()

class FinLeaf(FinInstrument):
    def price(self):
        mu = random.randint(1,100)
        return random.gauss(mu = mu, sigma = 10.)

class FiBasket(Basket):
    def base_class_get(self):
        return FinInstrument

    #def value(self, p: float) -> float:
    #    return p * sum(self.members.values())

class Trade(Traitable, Basketable, bucket_shape = BUCKET_SHAPE.DICT):
    name: str           = T(T.ID)
    booked_at: datetime = T()
    book1_name: str     = T()
    book2_name: str     = T()
    basket: FiBasket    = T()

    def members_qtys(self):
        return self.basket.members_qtys()

class Book(Traitable, Basketable, bucket_shape = BUCKET_SHAPE.SET):
    name: str                       = T(T.ID)
    customer_book: bool             = T()

    primary_trades: BucketDict      = RT()
    counterparty_trades: BucketDict = RT()

    def primary_trades_get(self) -> BucketDict:
        return BucketDict(data = { trade: 1. for trade in Trade.existing_instances_by_filter(f(book1_name = self.name)) })

    def counterparty_trades_get(self) -> BucketDict:
        return BucketDict(data = { trade: -1. for trade in Trade.existing_instances_by_filter(f(book2_name = self.name)) })

    def members_qtys(self):
        return itertools.chain(self.primary_trades.members_qtys(), self.counterparty_trades.members_qtys())

class Portfolio(Traitable, Basketable, bucket_shape = BUCKET_SHAPE.SET):
    name: str               = T(T.ID)

    portfolio_names: list   = T()
    book_names: list        = T()

    portfolios: Basket      = RT()
    books: Basket           = RT()

    def portfolios_get(self):
        basket = Basket(base_class = Portfolio, subclasses_allowed = False)
        for pname in self.portfolio_names:
            basket.add(Portfolio.existing_instance(name = pname))
        return basket

    def books_get(self):
        basket = Basket(base_class = Book, subclasses_allowed = False)
        for bname in self.book_names:
            basket.add(Book.existing_instance(name = bname))
        return basket

    def is_member(self, obj) -> bool:
        return self.portfolios.is_member(obj) or self.books.is_member(obj)

    def members_qtys(self):
        return itertools.chain(self.portfolios.members_qtys(), self.books.members_qtys())

if __name__ == '__main__':
    from core_10x.manual_tests.basketable_hierarchy_test import Portfolio, Book, Trade, FinInstrument, FinLeaf

    l1 = FinLeaf(name = 'Leaf 1')
    l2 = FinLeaf(name = 'Leaf 2')

    f1 = FinInstrument(
        _replace = True,
        name    = 'F1',
        leaves = {l1: 5., l2: -6.}
    )

    f2 = FinInstrument(name = 'F2')
    f3 = FinInstrument(name = 'F3')

    t1 = Trade(_replace = True,
        name        = 'T1',
        book1_name  = 'Book1',
        book2_name  = 'Book2',
        basket      = FiBasket(the_bucket = BucketDict(data = {f1: 10., f2: -2.}))
    )

    t2 = Trade(_replace = True,
        name        = 'T2',
        book1_name  = 'Book3',
        book2_name  = 'Book2',
        basket      = FiBasket(bucket = BucketDict(data = {f3: 4.}))
    )

    b1 = Book(_replace = True, name = 'Book1', customer_book = False)
    b2 = Book(_replace = True, name = 'Book2', customer_book = True)
    b3 = Book(_replace = True, name = 'Book3', customer_book = False)

    p1 = Portfolio(_replace = True, name    = 'P1', book_names = ['Book1'])
    p2 = Portfolio(_replace = True, name    = 'P2', book_names = ['Book2'])

    p = Portfolio(
        _replace = True,
        name    = 'Top',
        portfolio_names = ['P1', 'P2'],
        book_names = ['Book3']
    )

    basket1 = Basket(base_class = Book)
    p.contents(basket1)

    basket2 = Basket(base_class = Trade)
    p.contents(basket2)

    basket3 = Basket(base_class = FinInstrument)
    p.contents(basket3)


    # books = p.contents(Book)
    #
    # trades = p.contents(Trade)
    #
    # leaves = p.contents(FinInstrument, leaves = True)
    #
    # r = trades.value(1500.)




from core_10x.attic.backbone.backbone_traitable import RT, BackboneTraitable, T
from core_10x.attic.data_domain import DataDomain
from core_10x.environment_variables import EnvVars
from core_10x.resource import Resource
from core_10x.xnone import XNone


class ResourceSpec(BackboneTraitable):
    # fmt: off
    build_area: str         = T(T.ID)
    name: str               = T(T.ID)
    resource_class: type    = T()
    resource_kwargs: dict   = T()

    resource: Resource      = RT()
    # fmt: on

    def build_area_get(self) -> str:
        return EnvVars.sdlc_area

    def name_get(self) -> str:
        kwargs = self.resource_kwargs
        return kwargs.get(Resource.HOSTNAME_TAG, XNone)

    def resource_get(self) -> Resource:
        cls = self.resource_class
        assert issubclass(cls, Resource), f'{cls} is not a subclass of Resource'
        return cls.instance(**self.resource_kwargs)


class BoundDataDomain(BackboneTraitable):
    # fmt: off
    domain: DataDomain.__class__    = T(T.ID)
    build_area: str                 = T(T.ID)
    bound_categories: dict          = T()
    # fmt: on

    def build_area_get(self) -> str:
        return EnvVars.sdlc_area

    def resource(self, category: str, throw: bool = True) -> Resource:
        r_spec_id: str = self.bound_categories.get(category)
        if not r_spec_id:
            if throw:
                raise ValueError(f'{self.domain} - unknown category {category}')
            return None

        r_spec = ResourceSpec(_id=r_spec_id)
        return r_spec.resource



from typing import Any
from uuid import uuid4

import numpy
import pytest
from py10x_kernel import BTraitableProcessor, XCache

from core_10x.code_samples.person import Person as BasePerson
from core_10x.package_refactoring import PackageRefactoring
from core_10x.rc import RC_TRUE
from core_10x.trait_definition import T
from core_10x.ts_store import TsDuplicateKeyError


class EnhancedPerson(BasePerson):
    numpy_weight_lbs: Any = T()

    def numpy_weight_lbs_get(self):
        return numpy.float64(self.weight_lbs)


test_classes = {
    cls_name: type(
        cls_name,
        (EnhancedPerson,),
        {
            '__module__': __name__,
        },
        custom_collection=custom_collection
    )
    for custom_collection in (None,True)
    for cls_name in (f'Person#{uuid4().hex}{custom_collection and "#Custom" or ""}' for _ in range(2))
}

globals().update(test_classes)


@pytest.fixture(params=[True, False], ids=['custom_collection', 'default_collection'])
def ts_setup(ts_instance,request):
    Person, Person1 = list(test_classes.values())[request.param*2:][:2]
    c, c1 = [uuid4().hex if request.param else PackageRefactoring.find_class_id(cls) for cls in (Person,Person1)]

    with ts_instance:
        p = Person(first_name='John', last_name='Doe', _collection_name=c if request.param else None)
        p.set_values(age=30, weight_lbs=100)
        assert p._rev == 0
        assert p.save() == RC_TRUE
        assert p._rev == 1
        assert p.save() == RC_TRUE
        assert p._rev == 1
        assert p.id().collection_name == (c if request.param else None)


        p1 = Person1(first_name='Joe', last_name='Doe', _collection_name=c1 if request.param else None)
        p1.set_values(age=32, weight_lbs=200)
        assert p1.save()
        assert p1.age == 32
        assert p1._rev == 1
        assert p1.id().collection_name == (c1 if request.param else None)

    yield ts_instance, p, p1, c, c1

    # Cleanup
    XCache.clear()
    BTraitableProcessor.current().end_using()
    for cn in [c, c1]:
        ts_instance.delete_collection(cn)
    assert not {c, c1}.intersection(ts_instance.collection_names('.*'))


class TestTSStore:
    """Test class for TS Store functionality."""

    def test_collection(self, ts_setup):
        ts_store, _p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)
        assert collection is not None

    def test_save(self, ts_setup):
        ts_store, p, _p1, c, _c1= ts_setup

        collection = ts_store.collection(c)
        serialized_entity = p.serialize_object()
        _rev = collection.save(serialized_entity.copy())
        assert p._rev == _rev

        serialized_entity |= {'attr': {'nested': 'value'}}
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 1 == _rev
        assert collection.load(p.id().value) == serialized_entity

        # test that nested dictionary replaces rather than updates
        serialized_entity |= {'attr': {'nested1': 'value1'}}
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 2 == _rev
        # assert collection.load(_id) == serialized_entity|{'attr': {'nested': 'value', 'nested1': 'value1'}} #incorrect behavior
        assert collection.load(p.id().value) == serialized_entity

        # test that dots are not interpreted as nested fields at the top level
        serialized_entity |= {'attr.nested2': 'value2'}
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 3 == _rev
        assert collection.load(p.id().value) == serialized_entity

        # check that we can have dictionary keys starting with $
        serialized_entity |= {'foo': {'$foo': 1}}
        # with pytest.raises(pyts_store.errors.WriteError, match="Unrecognized expression '\\$foo',"): #incorrect behavior
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 4 == _rev
        assert collection.load(p.id().value) == serialized_entity

        # check that we can *not* have top level keys starting with $ (current behavior, not ideal, but prob. ok)
        serialized_entity |= {'$foo': 1}
        with pytest.raises(Exception, match='Use of undefined variable: foo'):
            _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 4 == _rev
        del serialized_entity['$foo']
        assert collection.load(p.id().value) == serialized_entity

        # test that dots are not interpreted as nested fields at the nested level
        serialized_entity |= {'attr': {'nested.value': 1}}
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 5 == _rev
        assert collection.load(p.id().value) == serialized_entity

        # check that we can unset fields
        del serialized_entity['attr']
        _rev = collection.save(serialized_entity.copy())
        serialized_entity |= dict(_rev=_rev)
        assert p._rev + 6 == _rev
        assert collection.load(p.id().value) == serialized_entity

    def test_delete_restore(self, ts_setup):
        ts_store, p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)
        serialized_entity = p.serialize_object()
        id_value = serialized_entity['_id']
        assert collection.delete(id_value)
        assert collection.load(id_value) is None
        # TODO: restore is not implemented so use save_new meanwhile
        # assert collection.restore(id_value)
        collection.save_new(serialized_entity)
        assert collection.load(id_value) == serialized_entity

    def test_find(self, ts_setup):
        ts_store, p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)
        result = collection.find()
        assert next(iter(result)) == p.serialize_object()
        assert list(result) == []

    def test_load(self, ts_setup):
        ts_store, p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)
        id_value = p.id().value
        result = collection.load(id_value)
        assert result == p.serialize_object()

    def test_delete(self, ts_setup):
        ts_store, _p, p1, _c, c1= ts_setup
        collection = ts_store.collection(c1)
        id_value = p1.id().value
        result = collection.delete(id_value)
        assert result
        result = collection.load(id_value)
        assert result is None

    def test_save_new_with_overwrite(self, ts_setup):
        """Test save_new with overwrite=True flag."""
        ts_store, p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)
        serialized_entity = p.serialize_object()
        id_value = serialized_entity['_id']

        # Try to save_new with overwrite=True on existing document
        result = collection.save_new(serialized_entity.copy(), overwrite=True)
        assert result == 1
        assert collection.load(id_value) == serialized_entity

    def test_save_new_with_set_operation(self, ts_setup):
        """Test save_new with $set MongoDB-style operation."""
        ts_store, _p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)

        # Create a new document with $set
        doc_id = 'test_doc_123'
        serialized_entity = {'$set': {'_id': doc_id, 'name': 'Test Document', 'value': 42}}

        result = collection.save_new(serialized_entity)
        assert result == 1

        loaded = collection.load(doc_id)
        assert loaded == serialized_entity['$set'] | {'_rev': 1}

    def test_save_new_duplicate_key_error(self, ts_setup):
        """Test that save_new raises TsDuplicateKeyError when inserting duplicate without overwrite."""
        ts_store, _p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)

        # Test 1: Duplicate without $set
        doc_id = 'duplicate_test_123'
        serialized_entity = {'_id': doc_id, 'name': 'First Document'}

        result = collection.save_new(serialized_entity)
        assert result == 1

        # Try to insert the same document again without overwrite (no $set)
        with pytest.raises(TsDuplicateKeyError, match=f'Duplicate key error collection.*dup key.*{doc_id}'):
            collection.save_new(serialized_entity, overwrite=False)

        # Test 2: Duplicate with $set
        doc_id2 = 'duplicate_test_456'
        serialized_entity2 = {'$set': {'_id': doc_id2, 'name': 'First Document with $set'}}

        result = collection.save_new(serialized_entity2)
        assert result == 1

        # Try to insert the same document again without overwrite (with $set)
        with pytest.raises(TsDuplicateKeyError, match=f'Duplicate key error collection.*dup key.*{doc_id2}'):
            collection.save_new(serialized_entity2, overwrite=False)

    def test_save_new_with_set_and_overwrite(self, ts_setup):
        """Test save_new with $set and overwrite=True."""
        ts_store, _p, _p1, c, _c1= ts_setup
        collection = ts_store.collection(c)

        doc_id = 'set_overwrite_test_123'
        # First insert
        serialized_entity1 = {'_id': doc_id, 'name': 'Original'}
        result = collection.save_new(serialized_entity1)
        assert result == 1

        # Update with $set and overwrite=True
        serialized_entity2 = {'$set': {'_id': doc_id, 'name': 'Updated', 'new_field': 'new_value'}}
        result = collection.save_new(serialized_entity2, overwrite=True)
        assert result == 1

        loaded = collection.load(doc_id)
        assert loaded['name'] == 'Updated'
        assert loaded['new_field'] == 'new_value'

    def test_ts_class_association_ts_uri_resolution(self, ts_instance):
        """Test that TsClassAssociation.ts_uri correctly resolves store URIs for classes and their subclasses."""
        from core_10x.py_class import PyClass
        from core_10x.traitable import NamedTsStore, T, Traitable, TsClassAssociation

        class Dummy1(Traitable):
            text: str = T(T.ID)

        class Dummy2(Traitable):
            text: str = T(T.ID)

        class Dummy3(Dummy2): ...

        dummy_uri1 = 'mongodb://localhost/dummy1'
        dummy_uri2 = 'mongodb://localhost/dummy2'

        with ts_instance:
            # Create and save NamedTsStore objects
            ns1 = NamedTsStore(logical_name='dummy1', uri=dummy_uri1, _replace=True)
            ns1.save()
            ns2 = NamedTsStore(logical_name='dummy2', uri=dummy_uri2, _replace=True)
            ns2.save()

            # Create and save TsClassAssociation objects
            name1 = PyClass.name(Dummy1)
            name2 = PyClass.name(Dummy2)
            TsClassAssociation(py_canonical_name=name1, ts_logical_name='dummy1', _replace=True).save()
            TsClassAssociation(py_canonical_name=name2, ts_logical_name='dummy2', _replace=True).save()

            # Class-specific association
            assert TsClassAssociation.ts_uri(Dummy1) == dummy_uri1

            # Subclass should inherit parent's association if it has none of its own
            assert TsClassAssociation.ts_uri(Dummy3) == dummy_uri2



if __name__ == '__main__':
    from datetime import date

    from core_10x.scenario import Scenario
    from core_10x.code_samples.person import Person, WEIGHT_QU

    Person.s_print = True

    p = Person(
        _replace = True,
        last_name   = 'Smith',
        first_name  = 'John',
        dob         = date(1970, 3, 12),
        weight      = 170.
    )
    print(f'No scenario: {p.weight}')

    fn = p.full_name
    with Scenario(fn):
        p.weight = 180.
        print(f'In Scenario({fn}): {p.weight}')

    print(f'No scenario: {p.weight}')
    with Scenario(fn):
        print(f'In Scenario({fn}): {p.weight}')

    with Scenario():
        print('In temp Scenario:')

        print(f'\tage = {p.age}')
        print(f'\tage = {p.age}')

        p.dob = date(1960, 3, 12)
        print(f'\tage = {p.age}')



from __future__ import annotations

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget


class RadioButton(Widget, i.RadioButton):
    __slots__ = '_button_group'
    s_component_class = rio_components.RadioButton

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'label' not in self._kwargs:
            self['label'] = self.get_children().pop(0)
        if 'value' not in self._kwargs:
            self['value'] = self._kwargs['label']
        self['on_select'] = self.on_select
        self['checked'] = False
        self._button_group = None

    def set_checked(self, checked: bool):
        self['checked'] = checked
        if self._button_group:
            self._button_group.on_change(self)

    def on_select(self):
        self.set_checked(not self['checked'])


class ButtonGroup(Widget, i.ButtonGroup):
    s_component_class = rio_components.GroupBox
    s_pass_children_in_kwargs = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def on_change(self, changed_button: RadioButton):
        if changed_button['checked']:
            for child in self.get_children():
                if child is not changed_button:
                    child['checked'] = False

    def add_button(self, button, id):
        if id < len(self.get_children()):
            self.get_children()[id] = button
        else:
            assert len(self.get_children()) == id
            self.get_children().append(button)
        button._button_group = self

    def button(self, id: int) -> RadioButton:
        return self.get_children()[id]

    def checked_id(self):
        # TODO: optimize
        for idx, button in enumerate(self.get_children()):
            if button['checked']:
                return idx
        return -1



from __future__ import annotations

import os
import random
from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QHeaderView

import core_10x.trait_definition as trait_definition
from core_10x.global_cache import cache
from core_10x.traitable import RC, RC_TRUE, RT, T, Traitable, Ui

from ui_10x.utils import ux, ux_warning, ux_success, ux_push_button

from ui_10x.table_view import TableView
from ui_10x.traitable_editor import TraitableEditor

if TYPE_CHECKING:
    from collections.abc import Generator


class CHAR_STATE:
    NONE        = ('lightgray',    'black')
    WRONG_POS   = ('white',        'black')
    BINGO       = ('darkgreen',    'white')

class GuessResult(Traitable):
    the_word: str    = RT(T.HIDDEN)
    guess: str       = RT(T.HIDDEN)
    num_chars: int   = RT(T.HIDDEN)

    demo_class: type = RT(T.HIDDEN)

    char_state: list = RT(T.HIDDEN)

    guessed_chars: list = RT(T.HIDDEN)
    current_char: int = RT(T.HIDDEN, default = 0)

    def guessed_chars_get(self) -> list:
        return list(self.num_chars * ' ')

    def set_current_char(self, c: str):
        i = self.current_char
        guessed_chars = self.guessed_chars
        guessed_chars[i] = c
        self.current_char = i + 1

    def delete_char(self):
        i = self.current_char
        if i:
            i -= 1
            guessed_chars = self.guessed_chars
            guessed_chars[i] = ' '
            self.current_char = i

    def accept_guess(self):
        self.guess = ''.join(self.guessed_chars)

    def num_chars_get(self) -> int:
        return len(self.the_word)

    def char_state_get(self) -> list:
        num_chars = self.num_chars
        res = [CHAR_STATE.NONE] * num_chars
        word = list(self.the_word)
        for i, c in enumerate(self.guess):
            try:
                pos = word.index(c)
                res[i] = CHAR_STATE.BINGO if pos == i else CHAR_STATE.WRONG_POS
                word[pos] = ''
            except ValueError:
                pass

        return res

    def guess_get(self) -> str:
        return self.num_chars * ' '

    @classmethod
    def _create_char_trait(cls, i: int):
        name = f'char_{i}'
        getter_name = f'{name}_get'
        sh_name = f'{name}_style_sheet'
        getter = lambda self: self.guess[i]
        getter.__name__ = getter_name
        sh_getter = lambda self: T.colors(*self.char_state[i])
        sh_getter.__name__ = sh_name
        setattr(cls, getter_name, getter)
        setattr(cls, sh_name, sh_getter)
        return (
            name,
            trait_definition.TraitDefinition(
                **{
                    trait_definition.NAME_TAG:      name,
                    trait_definition.DATATYPE_TAG:  str,
                    trait_definition.FLAGS_TAG:     T.RUNTIME,
                    trait_definition.UI_HINT_TAG:   Ui(label = f'{i+1}')
                }
            )
        )


# class Key:
#     def __init__(self, guess: GuessResult, c: str, stylesheet: str, callback):
#         self.guess = guess
#         self.c = c
#         self.stylesheet = stylesheet
#         self.callback = callback
#         self.w = None
#
#     # def widget(self, parent_w: ux.Widget) -> ux.PushButton:
#     def widget(self) -> ux.PushButton:
#         if not self.w:
#             self.w = b = ux.PushButton(self.c)
#             b.clicked_connect(lambda w: self.callback(w))
#             b.set_style_sheet(self.stylesheet)
#
#         return self.w
#
#
# class Keyboard:
#     s_keys = [
#         'qwertyuiop',
#         'asdfghjkl',
#         '*zxcvbnm<',
#     ]
#     s_special_chars = {
#         2:  {
#             0:                      ( 'background-color: green; color: white',  GuessResult.accept_guess ),
#             len(s_keys[2]) - 1:     ( 'background-color: red; color: white',    GuessResult.delete_char ),
#         }
#     }
#
#     def __init__(self, guess: GuessResult):
#         self.guess = guess
#         special_chars = self.s_special_chars
#         num_rows = len(self.s_keys)
#         self.keys = keys = [[]] * num_rows
#         for i, symbols in enumerate(self.s_keys):
#             for j, c in enumerate(symbols):
#                 sh = ''
#                 cb = GuessResult.set_current_char
#                 exc = special_chars.get(i)
#                 if exc:
#                     spec = exc.get(j)
#                     if spec:
#                         sh, cb = spec
#
#                 key = Key(guess, c, sh, cb)
#                 keys[i].append(key)
#
#     def widget(self) -> ux.Widget:
#         w = ux.Widget()
#         lay = ux.VBoxLayout()
#         w.set_layout(lay)
#
#         special_chars = self.s_special_chars
#         for i, keys in enumerate(self.s_keys):
#             hlay = ux.HBoxLayout()
#             for j, c in enumerate(keys):
#                 sh = ''
#                 cb = GuessResult.set_current_char
#                 exc = special_chars.get(i)
#                 if exc:
#                     spec = exc.get(j)
#                     if spec:
#                         sh, cb = spec
#
#                 b = ux_push_button(c, callback = cb)
#                 b.set_style_sheet(sh)
#                 avg_char_width = w.font_metrics().average_char_width()
#                 b.setMaximumWidth(3 * avg_char_width)
#
#                 hlay.add_widget(b)
#             lay.add_layout(hlay)
#
#
#         return w

class Game(Traitable):
    num_chars: int  = RT(T.HIDDEN, default = 5)

    the_word: str   = RT(T.HIDDEN)
    guess: str      = RT( ui_hint = Ui('>'))
    push: bool      = RT( ui_hint = Ui('Try', widget_type = Ui.WIDGET_TYPE.PUSH, right_label = True))

    count: int      = RT(T.HIDDEN, default = 6)
    current: int    = RT(T.HIDDEN, default = 0)

    guess_res_class: type   = RT(T.HIDDEN)
    attempts: list          = RT(T.HIDDEN)

    top_editor: TraitableEditor = RT()
    table: TableView        = RT()
    # keyboard: Keyboard      = RT()

    def the_word_get(self) -> str:
        return _GuessWordData.new_word(self.num_chars).upper()

    def guess_set(self, trait, value: str) -> RC:
        self.raw_set_value(trait, value.upper())
        return RC_TRUE

    def guess_res_class_get(self):
        num_chars = self.num_chars

        class Demo(GuessResult):
            @classmethod
            def own_trait_definitions(cls) -> Generator[tuple[str, trait_definition.TraitDefinition]]:
                yield from (tri for i in range(num_chars) if (tri := GuessResult._create_char_trait(i)))

        Demo.traits()
        return Demo

    def guess_result(self, guess: str):
        demo_class = self.guess_res_class
        return demo_class(the_word=self.the_word, guess=guess)

    def push_get(self) -> bool:
        return self.current < self.count

    def attempts_get(self) -> list:
        num_chars = self.num_chars
        return [self.guess_result(num_chars * ' ') for _ in range(self.count)]

    def push_action(self, editor):
        current = self.current
        guess = self.guess
        if len(guess) != self.num_chars:
            ux_warning(f'The word is {self.num_chars} characters long.')
            return

        if not _GuessWordData.word_exists(guess):
            ux_warning(f"Unfamiliar noun '{guess}'")
            return

        gr = self.guess_result(self.guess)
        attempts = self.attempts
        attempts[current] = gr
        self.table.render_traitable(current, None)

        if all(s == CHAR_STATE.BINGO for s in gr.char_state):
            ux_success('Great job - you got it!')
            return

        current += 1
        if current >= self.count:
            ux_warning('Unfortunately you have failed this time')

        self.current = current

    def top_editor_get(self) -> TraitableEditor:
        return TraitableEditor.editor(self)

    def table_get(self) -> TableView:
        table = TableView(self.attempts)
        hv = table.horizontalHeader()
        hv.setStretchLastSection(False)
        return table

    # def keyboard_get(self) -> Keyboard:
    #     return Keyboard(GuessResult())

    def widget(self):
        ux.init()
        w = ux.Widget()
        lay = ux.VBoxLayout()
        w.set_layout(lay)

        top = self.top_editor.row_layout()
        lay.add_layout(top)
        lay.add_widget(ux.separator())

        lay.add_widget(self.table)

        #lay.add_widget(ux.separator())
        #lay.add_widget(self.keyboard.widget())

        return w


class _GuessWordData:
    MODULE_NAME = '_guess_word_data'
    NUM_CHARS   = (5, 6)

    @classmethod
    def download_nouns(cls, filename: str):
        import nltk
        from nltk.corpus import wordnet as wn

        if os.path.exists(filename):
            return

        nltk.download('wordnet')
        noun_pool = [
            w.name() for s in wn.all_synsets('n') for w in s.lemmas() if len(w.name()) in cls.NUM_CHARS and w.name().isalpha() and w.name().islower()
        ]

        with open(filename, 'w') as f:
            print('noun_pool = {', file=f)
            for n in cls.NUM_CHARS:
                print(f'\t{n}: [', file=f)
                pool = [w for w in noun_pool if len(w) == n]
                for w in pool:
                    print(f'\t\t{w!r},', file=f)
                print('\t],', file=f)
            print('}', file=f)

    @classmethod
    @cache
    def noun_pool(cls) -> dict:
        import importlib
        import importlib.util
        from pathlib import Path

        package_name = cls.__module__.rsplit('.', 1)[0]
        spec = importlib.util.find_spec(package_name)
        assert spec.submodule_search_locations
        package_dir = Path(spec.submodule_search_locations[0])
        file_name = package_dir / f'{cls.MODULE_NAME}.py'
        #cls.download_nouns(file_name)

        module_name = f'{package_name}.{cls.MODULE_NAME}'
        m = importlib.import_module(module_name)
        noun_pool = getattr(m, 'NOUN_POOL', None)
        assert noun_pool and type(noun_pool) is dict, 'noun_pool must be a non-empty dict by num chars'
        return noun_pool

    @classmethod
    def word_exists(cls, w: str) -> bool:
        noun_pool = cls.noun_pool()
        n = len(w)
        words = noun_pool.get(n)
        return words and w.lower() in words

    @classmethod
    def new_word(cls, n: int) -> str:
        noun_pool = cls.noun_pool()
        words = noun_pool.get(n)
        if words is None:
            raise AssertionError(f'There are no {n}-letter words available')
        return random.choice(words)


if __name__ == '__main__':
    from core_10x.exec_control import INTERACTIVE
    from ui_10x.examples.guess_word import Game, _GuessWordData
    from ui_10x.utils import UxDialog

    with INTERACTIVE():
        game = Game()
        w = game.widget()

        d = UxDialog(w, title = f'You have {game.count} attempts to guess a word')
        d.exec()



from datetime import date

from core_10x.py_class import PyClass

from ui_10x.py_data_browser import PyDataBrowser
from ui_10x.utils import ux


class A:
    pass


class B:
    def __repr__(self):
        return f'class {self.__class__.__name__} stub'

data = {
    'Plain int': 33,
    'Boolean': True,
    'All': (False, True),
    'Just None': None,
    'Now': date.today(),
    'A simple list': [ 100, 'abracadabra', 375.89 ],
    'A simple dict': dict(_a = 10, _b = None, _c = 'Atlantic Ocean'),
    'Unknown': A(),
    'Unknown, but': B(),

    'Nested Dict': dict(
        person = dict(
            first_name = 'Mike',
            last_name = 'Fellows',
            age = 50,
            countries = ('USA', 'Japan', 'Korea'),
        ),
        regions = [
            'Europe',
            dict(
                North = [ 'Canada', 'USA', 'Mexico' ],
                Central = [ 'Nicaragua', 'Costa Rica', 'Equador' ],
                South = [ 'Brasil', 'Argentina' ],
            ),
            'Asia',
        ],
    ),
    # 'People':           Person.generateList( _cache_only = True )
}

def on_select(browser: PyDataBrowser, path: list, value, delims = ('/', '/')):
    path_text = PyDataBrowser.path_to_str(path, delims)
    return print(f"'{path_text}' -> {value}")


if __name__ == '__main__':
    ux.init()

    PyDataBrowser.show([10, 'a'], on_select = on_select)

    PyDataBrowser.edit(
        data,
        on_select = lambda browser, path, value: on_select(browser, path, value, delims = ('[]', '[]'))
    )

    PyDataBrowser.show(
        PyClass.class_names_by_module('core_10x'),
        on_select = lambda browser, path, value: on_select(browser, path, value, delims = ('.', ''))
    )



from datetime import date, datetime

from core_10x.rc import RC, RC_TRUE
from core_10x.xnone import XNone

from ui_10x.utils import UxDialog, ux


class PyDataBrowser:
    class Item(ux.TreeItem):
        def __init__(self, parent_node, key, value):
            super().__init__(parent_node)
            self.key = key
            self.value = value
            self.new_value = value

        def path(self) -> list:
            res = []
            node = self
            while node:
                res.insert(0, node.key)
                node = node.parent()

            return res

        def accept_edit(self, results):
            n = self.child_count()
            if not n:
                if self.value != self.new_value:
                    path = self.path()
                    x = results
                    for key in path[:-1]:
                        x = x[key]
                    x[path[-1]] = self.new_value
            else:
                for i in range(n):
                    child = self.child(i)
                    child.accept_edit(results)

    def __init__(
            self,
            data,
            on_select = None,
            edit = False,
            custom_editor = None,
            max_label_len = 50,
            max_value_len = 70
    ):
        """
        Browse any python data.
        :param data:
        :param on_select: a callable on item selection. If given, will call on_select( the_browser, path_to_item: list, item_value )
        :param edit: allow editing if True
        :param custom_editor: if given, is used for editing - fn( the_browser, item: PyDataBrowser: Item )
        :param max_label_len: max label width (if exceeds, gets truncated showing '...' and the end)
        :param max_value_len: max value width (if exceeds, gets truncated showing '...' and the end)
        Tooltip (full label/value, if possible) is shown as usual
        """
        self.data = data
        self.edit = edit
        self.custom_editor = custom_editor
        self.select_cb = on_select
        self.max_label_len = max_label_len
        self.max_value_len = max_value_len

    def fill(self, value, node: ux.TreeItem, label = True):
        if label:
            i = 0
            n = self.max_label_len
            tip = True
        else:
            i = 1
            n = self.max_value_len
            tip = isinstance(value, str)
        text = str(value)
        if len(text) > n:
            show = text[: n] + '...'
        else:
            show = text
        node.set_text(i, show)
        if tip:
            node.set_tool_tip(i, text)

    s_bool_vmap = {
        'True':     True,
        'False':    False,
        '1':        True,
        '0':        False
    }

    s_use_as_is = {
        type:           None,
        type(None):     None,
        type(XNone):    None,
        bool:           lambda x:   PyDataBrowser.s_bool_vmap.get(x, bool(x)),
        str:            lambda x:   x,
        int:            lambda x:   int(x),
        float:          lambda x:   float(x),
        datetime:       lambda x:   datetime.strptime(x, '%Y-%m-%d %H:%M:%S'),
        date:           lambda x:   datetime.strptime(x, '%Y-%m-%d').date(),
    }

    s_dict_handler  = lambda data: data.items()
    s_list_handler  = lambda data: enumerate(data)
    s_type_handlers = {
        tuple:      s_list_handler,
        list:       s_list_handler,
        dict:       s_dict_handler,
    }
    def create_tree(self, key, item, parent_node):
        node = self.Item(parent_node, key, item)
        self.fill(key, node, label = True)
        self.fill(item, node, label = False)

        dtype = type(item)
        if dtype in self.s_use_as_is:
            pass
        else:
            handler = self.s_type_handlers.get(dtype)
            if not handler:
                try:
                    text = item.__repr__()
                    node.set_text(1, text)
                except Exception:
                    node.set_text(1, '???')
                    return

                try:
                    item = item.to_dict()
                    handler = self.__class__.s_dict_handler
                except Exception:
                    return

            for key, value in handler(item):
                self.create_tree(key, value, node)

    def widget(self) -> ux.TreeWidget:
        self.tree = tree = ux.TreeWidget()
        tree.set_column_count(2)
        tree.set_header_labels(['Key', 'Value'])

        data = self.data
        dtype = type(data)
        handler = self.s_type_handlers.get(dtype)
        if not handler:
            data = { str(dtype): data }
            handler = self.s_dict_handler

        for key, item in handler(data):
            self.create_tree(key, item, tree)

        if len(data) == 1:
            tree.top_level_item(0).set_expanded(True)

        tree.resize_column_to_contents(0)
        tree.resize_column_to_contents(1)

        tree.item_expanded_connect(lambda item: tree.resize_column_to_contents(0))
        tree.item_pressed_connect(self.on_item_pressed)
        tree.item_changed_connect(self.on_item_changed)

        return tree

    def on_item_pressed(self, item: Item, col: int):
        if item:
            if col == 0 or not self.edit:
                if self.select_cb:
                    self.select_cb(self, item.path(), item.value)
                return

            if self.custom_editor:
                self.custom_editor(self, item)
            else:
                if not item.child_count():
                    self.generic_editor(item)

    def on_item_changed(self, item: Item, col: int):
        if col != 1:
            return

        dtype = type(item.value)
        from_str_handler = self.s_use_as_is.get(dtype)
        if from_str_handler:
            text = item.text(col)
            try:
                item.new_value = from_str_handler(text)
            except Exception:
                pass

    def generic_editor(self, item: Item):
        self.tree.open_persistent_editor(item, 1)
        self.tree.edit_item(item, 1)

    def accept_edit(self) -> RC:
        for i in range(self.tree.top_level_item_count()):
            item = self.tree.top_level_item(i)
            item.accept_edit(self.data)
        return RC_TRUE

    @classmethod
    def show(cls, data, on_select = None, title = '', w = 1000, h = 800):
        if not title:
            title = 'Python Data Browser'

        br = cls(data, on_select = on_select)
        d = UxDialog(
            br.widget(),
            title       = title,
            ok          = '',
            cancel      = 'Close',
            min_width   = w,
            min_height  = h
        )
        d.exec()

    @classmethod
    def edit(cls, data, on_select = None, custom_editor = None, title = '', w = 1000, h = 800) -> bool:
        if not title:
            title = 'Python Data Editor'

        br = cls(data, on_select = on_select, edit = True, custom_editor = custom_editor)
        d = UxDialog(
            br.widget(),
            accept_callback = br.accept_edit,
            title           = title,
            min_width       = w,
            min_height      = h
        )
        rc = d.exec()
        return bool(rc)

    @classmethod
    def path_to_str(cls, path: list, delims = ('/', '/')) -> str:
        res = []
        for i, key in enumerate(path):
            delim = delims[ isinstance(key, int) ]
            n = len(delim)
            if n == 0:
                continue

            if n == 1:
                key = f'{delim}{key}' if i > 0 else str(key)
            else:
                key = f'{delim[0]}{key}{delim[1]}'

            res.append(key)

        return ''.join(res)




import functools


class T:
    def __init__(self, name, get, *args):
        self.name = name
        self.get = get
        self.args = args

    def __get__(self, instance, owner):
        if not self.args:
            return self.get(instance)

        return functools.partial(self.get, self)

class E:
    def __init__(self, fn, ln):
        self.fn = fn
        self.ln = ln

    def full_name(self):
        return f'{self.ln}, {self.fn}'

    def w(self, q):
        return q * 10

    name    = T('name', full_name )
    weight  = T('weight', w, 10 )

if __name__ == '__main__':
    e = E('Sasha', 'Davidovich')

    print(e.name)
    print(e.weight(5))


from core_10x.testlib.fixtures import with_transactions
from core_10x.testlib.ts_tests import TestTSStore, ts_setup # collected by pytest
from core_10x.testlib.ts_store_transaction_tests import TestSaveIfChanged, TestTsStoreTransaction # collected by pytest
from infra_10x.mongodb_store import MongoCollection, MongoStore


def test_mongo_parse_uri_round_trip():
    uri = 'mongodb://user:pass@localhost:27017/testdb?ssl=false&serverSelectionTimeoutMS=5000'
    args = MongoStore.parse_uri(uri)

    assert args[MongoStore.HOSTNAME_TAG] == 'localhost'
    assert args[MongoStore.DBNAME_TAG] == 'testdb'
    assert args[MongoStore.USERNAME_TAG] == 'user'
    assert args[MongoStore.PASSWORD_TAG] == 'pass'
    assert args['port'] == 27017
    # Options are driver-dependent; we only require that custom options are propagated.
    assert 'serverSelectionTimeoutMS' in args


def test_filter_and_pipeline_equivalence():
    from py10x_infra import MongoCollectionHelper
    from infra_10x.testlib.mongo_collection_helper import MongoCollectionHelperStub

    serialized_traitable = dict(_id='AAAA', _rev=10, name='test', age=60)

    data1 = dict(serialized_traitable)
    pipeline1: list = []
    filter1: dict = {}
    MongoCollectionHelperStub.prepare_filter_and_pipeline(data1, filter1, pipeline1)

    data2 = dict(serialized_traitable)
    pipeline2: list = []
    filter2: dict = {}
    MongoCollectionHelper.prepare_filter_and_pipeline(data2, filter2, pipeline2)

    assert filter1 == filter2
    assert pipeline1 == pipeline2
    assert data1 == data2



from core_10x.named_constant import NamedConstant
from core_10x.py_class import PyClass


class TS_STORE_TYPE(NamedConstant):
    """
    All known TsStore subclasses must be "registered" here.
    """

    MONGODB = 'infra_10x.mongodb_store.MongoStore'
    ...

    @classmethod
    def ts_store_class(cls, uri_protocol: str):
        """
        example: TS_STORE_TYPE.ts_store_class('mongodb')
        """
        symbol = cls.s_dir.get(uri_protocol.upper())
        if symbol is None:
            raise ValueError(f'Unknown URI protocol: {uri_protocol}')

        ts_class = PyClass.find(symbol.value)
        if not ts_class:
            raise TypeError(f'{symbol.value} - unknown TsStore class')

        return ts_class



from py10x_kernel import NODE_TYPE, BasicNode

from core_10x.xnone import XNone  # -- keep

n1 = BasicNode.create(NODE_TYPE.BASIC_GRAPH)
n1.assign('whatever')

n2 = BasicNode.create(NODE_TYPE.BASIC_GRAPH)
n2.set(100)

n2.add_parent(n1)

n2.set('Asya')



if __name__ == '__main__':
    from core_10x.code_samples.person import Person

    ppl_stored = [
        Person(last_name='Davidovich', first_name='Sasha', weight_lbs=170, _replace=True),
        Person(last_name='Pevzner', first_name='Ilya', weight_lbs=200, _replace=True),
        Person(last_name='Lesin', first_name='Alex', weight_lbs=190, _replace=True),
        Person(last_name='Smith', first_name='John', weight_lbs=180, _replace=True),
    ]

    id_traits = ('last_name', 'first_name')
    existing_id_traits = [{name: p.get_value(name) for name in id_traits} for p in ppl_stored]

    ppl_found = [Person.existing_instance(**id_values) for id_values in existing_id_traits]

    assert ppl_found == ppl_stored

    p1 = Person.existing_instance(last_name='Smith', first_name='John')
    assert p1

    p2 = Person.existing_instance(last_name='Smith', first_name='Josh', _throw=False)
    assert not p2

    id1 = ppl_found[0].id()
    p3 = Person.existing_instance_by_id(_id=id1)

    id1_val = id1.value
    p4 = Person.existing_instance_by_id(_id_value=id1_val)
    assert p3 == p4

    id2_val = 'Smith|Josh'
    p5 = Person.existing_instance_by_id(_id_value=id2_val, _throw=False)
    assert not p5
